"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

import builtins
import collections.abc
from . import common_pb2
from . import datasource_pb2
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
from . import spikesorter_pb2
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _TrainPhase:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _TrainPhaseEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_TrainPhase.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    TRAIN_PHASE_0: _TrainPhase.ValueType  # 0
    TRAIN_PHASE_1: _TrainPhase.ValueType  # 1
    TRAIN_PHASE_2: _TrainPhase.ValueType  # 2
    TRAIN_PHASE_3: _TrainPhase.ValueType  # 3

class TrainPhase(_TrainPhase, metaclass=_TrainPhaseEnumTypeWrapper): ...

TRAIN_PHASE_0: TrainPhase.ValueType  # 0
TRAIN_PHASE_1: TrainPhase.ValueType  # 1
TRAIN_PHASE_2: TrainPhase.ValueType  # 2
TRAIN_PHASE_3: TrainPhase.ValueType  # 3
global___TrainPhase = TrainPhase

class _LossType:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _LossTypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_LossType.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    LOSS_TYPE_MSE: _LossType.ValueType  # 0
    LOSS_TYPE_WMSE: _LossType.ValueType  # 1
    LOSS_TYPE_CROSS_ENTROPY: _LossType.ValueType  # 2

class LossType(_LossType, metaclass=_LossTypeEnumTypeWrapper): ...

LOSS_TYPE_MSE: LossType.ValueType  # 0
LOSS_TYPE_WMSE: LossType.ValueType  # 1
LOSS_TYPE_CROSS_ENTROPY: LossType.ValueType  # 2
global___LossType = LossType

class _SpkSortTrainerNetworkSize:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _SpkSortTrainerNetworkSizeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_SpkSortTrainerNetworkSize.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    SST_SIZE_DEFAULT: _SpkSortTrainerNetworkSize.ValueType  # 0
    SST_SIZE_ARB: _SpkSortTrainerNetworkSize.ValueType  # 1
    SST_SIZE_SMALL: _SpkSortTrainerNetworkSize.ValueType  # 2
    SST_SIZE_NOMINAL: _SpkSortTrainerNetworkSize.ValueType  # 3
    SST_SIZE_LARGE: _SpkSortTrainerNetworkSize.ValueType  # 4

class SpkSortTrainerNetworkSize(_SpkSortTrainerNetworkSize, metaclass=_SpkSortTrainerNetworkSizeEnumTypeWrapper): ...

SST_SIZE_DEFAULT: SpkSortTrainerNetworkSize.ValueType  # 0
SST_SIZE_ARB: SpkSortTrainerNetworkSize.ValueType  # 1
SST_SIZE_SMALL: SpkSortTrainerNetworkSize.ValueType  # 2
SST_SIZE_NOMINAL: SpkSortTrainerNetworkSize.ValueType  # 3
SST_SIZE_LARGE: SpkSortTrainerNetworkSize.ValueType  # 4
global___SpkSortTrainerNetworkSize = SpkSortTrainerNetworkSize

class _SpkSortTrainerNeighborPattern:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _SpkSortTrainerNeighborPatternEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_SpkSortTrainerNeighborPattern.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    SST_PATTERN_DEFAULT: _SpkSortTrainerNeighborPattern.ValueType  # 0
    SST_PATTERN_ARB: _SpkSortTrainerNeighborPattern.ValueType  # 1
    SST_PATTERN_DIST: _SpkSortTrainerNeighborPattern.ValueType  # 2
    SST_PATTERN_SPATIAL_TIGHT: _SpkSortTrainerNeighborPattern.ValueType  # 3
    SST_PATTERN_SPATIAL_LOOSE: _SpkSortTrainerNeighborPattern.ValueType  # 4

class SpkSortTrainerNeighborPattern(_SpkSortTrainerNeighborPattern, metaclass=_SpkSortTrainerNeighborPatternEnumTypeWrapper): ...

SST_PATTERN_DEFAULT: SpkSortTrainerNeighborPattern.ValueType  # 0
SST_PATTERN_ARB: SpkSortTrainerNeighborPattern.ValueType  # 1
SST_PATTERN_DIST: SpkSortTrainerNeighborPattern.ValueType  # 2
SST_PATTERN_SPATIAL_TIGHT: SpkSortTrainerNeighborPattern.ValueType  # 3
SST_PATTERN_SPATIAL_LOOSE: SpkSortTrainerNeighborPattern.ValueType  # 4
global___SpkSortTrainerNeighborPattern = SpkSortTrainerNeighborPattern

class _SpkSortTrainerNetworkRelease:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _SpkSortTrainerNetworkReleaseEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_SpkSortTrainerNetworkRelease.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    SST_NET_RELEASE_ANY: _SpkSortTrainerNetworkRelease.ValueType  # 0
    SST_NET_RELEASE_DEV: _SpkSortTrainerNetworkRelease.ValueType  # 1
    SST_NET_RELEASE_DEV_MOD: _SpkSortTrainerNetworkRelease.ValueType  # 2
    SST_NET_RELEASE_PROD: _SpkSortTrainerNetworkRelease.ValueType  # 3
    SST_NET_RELEASE_PROD_MOD: _SpkSortTrainerNetworkRelease.ValueType  # 4
    SST_NET_RELEASE_SANDBOX: _SpkSortTrainerNetworkRelease.ValueType  # 5
    SST_NET_RELEASE_SANDBOX_MOD: _SpkSortTrainerNetworkRelease.ValueType  # 6

class SpkSortTrainerNetworkRelease(_SpkSortTrainerNetworkRelease, metaclass=_SpkSortTrainerNetworkReleaseEnumTypeWrapper): ...

SST_NET_RELEASE_ANY: SpkSortTrainerNetworkRelease.ValueType  # 0
SST_NET_RELEASE_DEV: SpkSortTrainerNetworkRelease.ValueType  # 1
SST_NET_RELEASE_DEV_MOD: SpkSortTrainerNetworkRelease.ValueType  # 2
SST_NET_RELEASE_PROD: SpkSortTrainerNetworkRelease.ValueType  # 3
SST_NET_RELEASE_PROD_MOD: SpkSortTrainerNetworkRelease.ValueType  # 4
SST_NET_RELEASE_SANDBOX: SpkSortTrainerNetworkRelease.ValueType  # 5
SST_NET_RELEASE_SANDBOX_MOD: SpkSortTrainerNetworkRelease.ValueType  # 6
global___SpkSortTrainerNetworkRelease = SpkSortTrainerNetworkRelease

class _WrangleMode:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _WrangleModeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_WrangleMode.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    WRANGLE_MODE_IMPORT_NNX_VEX_SPIKES_SIM: _WrangleMode.ValueType  # 0
    WRANGLE_MODE_MERGE_FIT_VIEW: _WrangleMode.ValueType  # 1
    WRANGLE_MODE_IMPORT_KILOSORT_SPIKES: _WrangleMode.ValueType  # 2

class WrangleMode(_WrangleMode, metaclass=_WrangleModeEnumTypeWrapper): ...

WRANGLE_MODE_IMPORT_NNX_VEX_SPIKES_SIM: WrangleMode.ValueType  # 0
WRANGLE_MODE_MERGE_FIT_VIEW: WrangleMode.ValueType  # 1
WRANGLE_MODE_IMPORT_KILOSORT_SPIKES: WrangleMode.ValueType  # 2
global___WrangleMode = WrangleMode

@typing.final
class SpkSortTrainerProfileReq(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DSRC_FIELD_NUMBER: builtins.int
    MODELUID_FIELD_NUMBER: builtins.int
    OUTPATH_FIELD_NUMBER: builtins.int
    ISBLOCK_FIELD_NUMBER: builtins.int
    PHASE1TRAIN_FIELD_NUMBER: builtins.int
    POSSCALE_FIELD_NUMBER: builtins.int
    EPS_FIELD_NUMBER: builtins.int
    MINPTS_FIELD_NUMBER: builtins.int
    modelUID: builtins.str
    outPath: builtins.str
    isBlock: builtins.bool
    phase1Train: builtins.bool
    posScale: builtins.float
    eps: builtins.float
    minPts: builtins.int
    @property
    def dsrc(self) -> common_pb2.FileDescriptor: ...
    def __init__(
        self,
        *,
        dsrc: common_pb2.FileDescriptor | None = ...,
        modelUID: builtins.str | None = ...,
        outPath: builtins.str | None = ...,
        isBlock: builtins.bool | None = ...,
        phase1Train: builtins.bool | None = ...,
        posScale: builtins.float | None = ...,
        eps: builtins.float | None = ...,
        minPts: builtins.int | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["dsrc", b"dsrc", "eps", b"eps", "isBlock", b"isBlock", "minPts", b"minPts", "modelUID", b"modelUID", "outPath", b"outPath", "phase1Train", b"phase1Train", "posScale", b"posScale"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["dsrc", b"dsrc", "eps", b"eps", "isBlock", b"isBlock", "minPts", b"minPts", "modelUID", b"modelUID", "outPath", b"outPath", "phase1Train", b"phase1Train", "posScale", b"posScale"]) -> None: ...

global___SpkSortTrainerProfileReq = SpkSortTrainerProfileReq

@typing.final
class SpkSortTrainerDeleteModelReq(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    RELEASE_FIELD_NUMBER: builtins.int
    ISDELETENAMED_FIELD_NUMBER: builtins.int
    release: global___SpkSortTrainerNetworkRelease.ValueType
    isDeleteNamed: builtins.bool
    """!!TODO will need to complete this as needed."""
    def __init__(
        self,
        *,
        release: global___SpkSortTrainerNetworkRelease.ValueType | None = ...,
        isDeleteNamed: builtins.bool | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["isDeleteNamed", b"isDeleteNamed", "release", b"release"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["isDeleteNamed", b"isDeleteNamed", "release", b"release"]) -> None: ...

global___SpkSortTrainerDeleteModelReq = SpkSortTrainerDeleteModelReq

@typing.final
class SpkSortNetworkParams(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    INPUTFEATURETYPE_FIELD_NUMBER: builtins.int
    NUMFEATURES_FIELD_NUMBER: builtins.int
    LATENTDIM_FIELD_NUMBER: builtins.int
    inputFeatureType: spikesorter_pb2.InputFeatureType.ValueType
    numFeatures: builtins.int
    latentDim: builtins.int
    def __init__(
        self,
        *,
        inputFeatureType: spikesorter_pb2.InputFeatureType.ValueType | None = ...,
        numFeatures: builtins.int | None = ...,
        latentDim: builtins.int | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["inputFeatureType", b"inputFeatureType", "latentDim", b"latentDim", "numFeatures", b"numFeatures"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["inputFeatureType", b"inputFeatureType", "latentDim", b"latentDim", "numFeatures", b"numFeatures"]) -> None: ...

global___SpkSortNetworkParams = SpkSortNetworkParams

@typing.final
class DeepClustererCreationParams(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ENCDIMS_FIELD_NUMBER: builtins.int
    ENCACT_FIELD_NUMBER: builtins.int
    DECDIMS_FIELD_NUMBER: builtins.int
    DECACT_FIELD_NUMBER: builtins.int
    DROPOUTRATE_FIELD_NUMBER: builtins.int
    DROPOUTVAL_FIELD_NUMBER: builtins.int
    BATCHNORM_FIELD_NUMBER: builtins.int
    BATCHNORMMOMENTUM_FIELD_NUMBER: builtins.int
    BATCHNORMBEFOREACTIVATION_FIELD_NUMBER: builtins.int
    BATCHNORMLATENT_FIELD_NUMBER: builtins.int
    BATCHNORMRECONSTRUCTION_FIELD_NUMBER: builtins.int
    ACTRECONSTRUCTION_FIELD_NUMBER: builtins.int
    ACTLATENT_FIELD_NUMBER: builtins.int
    SAVEPATH_FIELD_NUMBER: builtins.int
    BASENAME_FIELD_NUMBER: builtins.int
    BASESEED_FIELD_NUMBER: builtins.int
    PROJDIMS_FIELD_NUMBER: builtins.int
    PROJACT_FIELD_NUMBER: builtins.int
    SEED_FIELD_NUMBER: builtins.int
    encAct: builtins.str
    decAct: builtins.str
    dropoutRate: builtins.float
    dropoutVal: builtins.float
    batchNorm: builtins.bool
    batchNormMomentum: builtins.float
    batchNormBeforeActivation: builtins.bool
    batchNormLatent: builtins.bool
    batchNormReconstruction: builtins.bool
    actReconstruction: builtins.str
    actLatent: builtins.str
    savePath: builtins.str
    baseName: builtins.str
    baseSeed: builtins.int
    projAct: builtins.str
    seed: builtins.int
    @property
    def encDims(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    @property
    def decDims(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    @property
    def projDims(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    def __init__(
        self,
        *,
        encDims: collections.abc.Iterable[builtins.int] | None = ...,
        encAct: builtins.str | None = ...,
        decDims: collections.abc.Iterable[builtins.int] | None = ...,
        decAct: builtins.str | None = ...,
        dropoutRate: builtins.float | None = ...,
        dropoutVal: builtins.float | None = ...,
        batchNorm: builtins.bool | None = ...,
        batchNormMomentum: builtins.float | None = ...,
        batchNormBeforeActivation: builtins.bool | None = ...,
        batchNormLatent: builtins.bool | None = ...,
        batchNormReconstruction: builtins.bool | None = ...,
        actReconstruction: builtins.str | None = ...,
        actLatent: builtins.str | None = ...,
        savePath: builtins.str | None = ...,
        baseName: builtins.str | None = ...,
        baseSeed: builtins.int | None = ...,
        projDims: collections.abc.Iterable[builtins.int] | None = ...,
        projAct: builtins.str | None = ...,
        seed: builtins.int | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["actLatent", b"actLatent", "actReconstruction", b"actReconstruction", "baseName", b"baseName", "baseSeed", b"baseSeed", "batchNorm", b"batchNorm", "batchNormBeforeActivation", b"batchNormBeforeActivation", "batchNormLatent", b"batchNormLatent", "batchNormMomentum", b"batchNormMomentum", "batchNormReconstruction", b"batchNormReconstruction", "decAct", b"decAct", "dropoutRate", b"dropoutRate", "dropoutVal", b"dropoutVal", "encAct", b"encAct", "projAct", b"projAct", "savePath", b"savePath", "seed", b"seed"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["actLatent", b"actLatent", "actReconstruction", b"actReconstruction", "baseName", b"baseName", "baseSeed", b"baseSeed", "batchNorm", b"batchNorm", "batchNormBeforeActivation", b"batchNormBeforeActivation", "batchNormLatent", b"batchNormLatent", "batchNormMomentum", b"batchNormMomentum", "batchNormReconstruction", b"batchNormReconstruction", "decAct", b"decAct", "decDims", b"decDims", "dropoutRate", b"dropoutRate", "dropoutVal", b"dropoutVal", "encAct", b"encAct", "encDims", b"encDims", "projAct", b"projAct", "projDims", b"projDims", "savePath", b"savePath", "seed", b"seed"]) -> None: ...

global___DeepClustererCreationParams = DeepClustererCreationParams

@typing.final
class SpkSortModelBuilderParams(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DEEPCLUSTERERCREATIONPARAMS_FIELD_NUMBER: builtins.int
    INPUTFEATURETYPE_FIELD_NUMBER: builtins.int
    NUMFEATURES_FIELD_NUMBER: builtins.int
    WFMRESAMPPTS_FIELD_NUMBER: builtins.int
    MODELUID_FIELD_NUMBER: builtins.int
    inputFeatureType: spikesorter_pb2.InputFeatureType.ValueType
    numFeatures: builtins.int
    wfmResampPts: builtins.int
    modelUID: builtins.str
    @property
    def deepClustererCreationParams(self) -> global___DeepClustererCreationParams: ...
    def __init__(
        self,
        *,
        deepClustererCreationParams: global___DeepClustererCreationParams | None = ...,
        inputFeatureType: spikesorter_pb2.InputFeatureType.ValueType | None = ...,
        numFeatures: builtins.int | None = ...,
        wfmResampPts: builtins.int | None = ...,
        modelUID: builtins.str | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["deepClustererCreationParams", b"deepClustererCreationParams", "inputFeatureType", b"inputFeatureType", "modelUID", b"modelUID", "numFeatures", b"numFeatures", "wfmResampPts", b"wfmResampPts"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["deepClustererCreationParams", b"deepClustererCreationParams", "inputFeatureType", b"inputFeatureType", "modelUID", b"modelUID", "numFeatures", b"numFeatures", "wfmResampPts", b"wfmResampPts"]) -> None: ...

global___SpkSortModelBuilderParams = SpkSortModelBuilderParams

@typing.final
class SpkSortModelTrainParams(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DSETTRAIN_FIELD_NUMBER: builtins.int
    TRAINPHASE_FIELD_NUMBER: builtins.int
    NUMEPOCHS_FIELD_NUMBER: builtins.int
    MINIBATCHSIZE_FIELD_NUMBER: builtins.int
    BASEALPHA_FIELD_NUMBER: builtins.int
    ALPHADECAYRATE_FIELD_NUMBER: builtins.int
    ALPHADECAYSTEP_FIELD_NUMBER: builtins.int
    LOSSRECON_FIELD_NUMBER: builtins.int
    LOSSCLASS_FIELD_NUMBER: builtins.int
    LOSSCLUST_FIELD_NUMBER: builtins.int
    L1_FIELD_NUMBER: builtins.int
    L2_FIELD_NUMBER: builtins.int
    NUMLOGS_FIELD_NUMBER: builtins.int
    ISSAVEMODEL_FIELD_NUMBER: builtins.int
    TEMPERATURE_FIELD_NUMBER: builtins.int
    BRAINREGION_FIELD_NUMBER: builtins.int
    PROBEID_FIELD_NUMBER: builtins.int
    NUMGTNEURONS_FIELD_NUMBER: builtins.int
    trainPhase: global___TrainPhase.ValueType
    numEpochs: builtins.int
    minibatchSize: builtins.int
    baseAlpha: builtins.float
    alphaDecayRate: builtins.float
    alphaDecayStep: builtins.int
    lossRecon: builtins.bool
    lossClass: builtins.bool
    lossClust: builtins.bool
    l1: builtins.float
    l2: builtins.float
    numLogs: builtins.int
    isSaveModel: builtins.bool
    temperature: builtins.float
    brainRegion: builtins.str
    probeID: builtins.str
    numGtNeurons: builtins.int
    @property
    def dsetTrain(self) -> datasource_pb2.DataSourceSetSaveRequest: ...
    def __init__(
        self,
        *,
        dsetTrain: datasource_pb2.DataSourceSetSaveRequest | None = ...,
        trainPhase: global___TrainPhase.ValueType | None = ...,
        numEpochs: builtins.int | None = ...,
        minibatchSize: builtins.int | None = ...,
        baseAlpha: builtins.float | None = ...,
        alphaDecayRate: builtins.float | None = ...,
        alphaDecayStep: builtins.int | None = ...,
        lossRecon: builtins.bool | None = ...,
        lossClass: builtins.bool | None = ...,
        lossClust: builtins.bool | None = ...,
        l1: builtins.float | None = ...,
        l2: builtins.float | None = ...,
        numLogs: builtins.int | None = ...,
        isSaveModel: builtins.bool | None = ...,
        temperature: builtins.float | None = ...,
        brainRegion: builtins.str | None = ...,
        probeID: builtins.str | None = ...,
        numGtNeurons: builtins.int | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["alphaDecayRate", b"alphaDecayRate", "alphaDecayStep", b"alphaDecayStep", "baseAlpha", b"baseAlpha", "brainRegion", b"brainRegion", "dsetTrain", b"dsetTrain", "isSaveModel", b"isSaveModel", "l1", b"l1", "l2", b"l2", "lossClass", b"lossClass", "lossClust", b"lossClust", "lossRecon", b"lossRecon", "minibatchSize", b"minibatchSize", "numEpochs", b"numEpochs", "numGtNeurons", b"numGtNeurons", "numLogs", b"numLogs", "probeID", b"probeID", "temperature", b"temperature", "trainPhase", b"trainPhase"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["alphaDecayRate", b"alphaDecayRate", "alphaDecayStep", b"alphaDecayStep", "baseAlpha", b"baseAlpha", "brainRegion", b"brainRegion", "dsetTrain", b"dsetTrain", "isSaveModel", b"isSaveModel", "l1", b"l1", "l2", b"l2", "lossClass", b"lossClass", "lossClust", b"lossClust", "lossRecon", b"lossRecon", "minibatchSize", b"minibatchSize", "numEpochs", b"numEpochs", "numGtNeurons", b"numGtNeurons", "numLogs", b"numLogs", "probeID", b"probeID", "temperature", b"temperature", "trainPhase", b"trainPhase"]) -> None: ...

global___SpkSortModelTrainParams = SpkSortModelTrainParams

@typing.final
class SpkSortTrainerNewReq(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PARAMSMODELBUILDER_FIELD_NUMBER: builtins.int
    PARAMSTRAINPHASE1_FIELD_NUMBER: builtins.int
    PARAMSTRAINPHASE2_FIELD_NUMBER: builtins.int
    PARAMSTRAINPHASE3_FIELD_NUMBER: builtins.int
    ISBLOCK_FIELD_NUMBER: builtins.int
    isBlock: builtins.bool
    @property
    def paramsModelBuilder(self) -> global___SpkSortModelBuilderParams: ...
    @property
    def paramsTrainPhase1(self) -> global___SpkSortModelTrainParams: ...
    @property
    def paramsTrainPhase2(self) -> global___SpkSortModelTrainParams: ...
    @property
    def paramsTrainPhase3(self) -> global___SpkSortModelTrainParams: ...
    def __init__(
        self,
        *,
        paramsModelBuilder: global___SpkSortModelBuilderParams | None = ...,
        paramsTrainPhase1: global___SpkSortModelTrainParams | None = ...,
        paramsTrainPhase2: global___SpkSortModelTrainParams | None = ...,
        paramsTrainPhase3: global___SpkSortModelTrainParams | None = ...,
        isBlock: builtins.bool | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["isBlock", b"isBlock", "paramsModelBuilder", b"paramsModelBuilder", "paramsTrainPhase1", b"paramsTrainPhase1", "paramsTrainPhase2", b"paramsTrainPhase2", "paramsTrainPhase3", b"paramsTrainPhase3"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["isBlock", b"isBlock", "paramsModelBuilder", b"paramsModelBuilder", "paramsTrainPhase1", b"paramsTrainPhase1", "paramsTrainPhase2", b"paramsTrainPhase2", "paramsTrainPhase3", b"paramsTrainPhase3"]) -> None: ...

global___SpkSortTrainerNewReq = SpkSortTrainerNewReq

@typing.final
class SpkSortLearnHyperParamsReq(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MODELUID_FIELD_NUMBER: builtins.int
    ISBLOCK_FIELD_NUMBER: builtins.int
    ISPOSSCALE_FIELD_NUMBER: builtins.int
    ISDBSCAN_FIELD_NUMBER: builtins.int
    DSRC_FIELD_NUMBER: builtins.int
    modelUID: builtins.str
    isBlock: builtins.bool
    isPosScale: builtins.bool
    isDBScan: builtins.bool
    @property
    def dsrc(self) -> common_pb2.FileDescriptor: ...
    def __init__(
        self,
        *,
        modelUID: builtins.str | None = ...,
        isBlock: builtins.bool | None = ...,
        isPosScale: builtins.bool | None = ...,
        isDBScan: builtins.bool | None = ...,
        dsrc: common_pb2.FileDescriptor | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["dsrc", b"dsrc", "isBlock", b"isBlock", "isDBScan", b"isDBScan", "isPosScale", b"isPosScale", "modelUID", b"modelUID"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["dsrc", b"dsrc", "isBlock", b"isBlock", "isDBScan", b"isDBScan", "isPosScale", b"isPosScale", "modelUID", b"modelUID"]) -> None: ...

global___SpkSortLearnHyperParamsReq = SpkSortLearnHyperParamsReq

@typing.final
class SpkSortTrainerDatasetParam(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NIPATHTRAIN_FIELD_NUMBER: builtins.int
    NIBASENAMETRAIN_FIELD_NUMBER: builtins.int
    NBHDDIM_FIELD_NUMBER: builtins.int
    RADIUSUM_FIELD_NUMBER: builtins.int
    MAXNUMSAMPLESTEST_FIELD_NUMBER: builtins.int
    NETWORKMASKINGFRAC_FIELD_NUMBER: builtins.int
    NETWORKINITWEIGHTSSTDDEV_FIELD_NUMBER: builtins.int
    NETWORKINITBIASSTDDEV_FIELD_NUMBER: builtins.int
    NETWORKL1LAMBDA_FIELD_NUMBER: builtins.int
    NETWORKHIDDENLAYERSHAPERELINPUTDIM_FIELD_NUMBER: builtins.int
    NETWORKNUMFEATURES2_FIELD_NUMBER: builtins.int
    niPathTrain: builtins.str
    niBaseNameTrain: builtins.str
    NbhdDim: builtins.int
    radiusUm: builtins.float
    maxNumSamplesTest: builtins.int
    networkMaskingFrac: builtins.float
    networkInitWeightsStdDev: builtins.float
    networkInitBiasStdDev: builtins.float
    networkL1lambda: builtins.float
    networkNumFeatures2: builtins.int
    @property
    def networkHiddenLayerShapeRelInputDim(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.float]: ...
    def __init__(
        self,
        *,
        niPathTrain: builtins.str | None = ...,
        niBaseNameTrain: builtins.str | None = ...,
        NbhdDim: builtins.int | None = ...,
        radiusUm: builtins.float | None = ...,
        maxNumSamplesTest: builtins.int | None = ...,
        networkMaskingFrac: builtins.float | None = ...,
        networkInitWeightsStdDev: builtins.float | None = ...,
        networkInitBiasStdDev: builtins.float | None = ...,
        networkL1lambda: builtins.float | None = ...,
        networkHiddenLayerShapeRelInputDim: collections.abc.Iterable[builtins.float] | None = ...,
        networkNumFeatures2: builtins.int | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["NbhdDim", b"NbhdDim", "maxNumSamplesTest", b"maxNumSamplesTest", "networkInitBiasStdDev", b"networkInitBiasStdDev", "networkInitWeightsStdDev", b"networkInitWeightsStdDev", "networkL1lambda", b"networkL1lambda", "networkMaskingFrac", b"networkMaskingFrac", "networkNumFeatures2", b"networkNumFeatures2", "niBaseNameTrain", b"niBaseNameTrain", "niPathTrain", b"niPathTrain", "radiusUm", b"radiusUm"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["NbhdDim", b"NbhdDim", "maxNumSamplesTest", b"maxNumSamplesTest", "networkHiddenLayerShapeRelInputDim", b"networkHiddenLayerShapeRelInputDim", "networkInitBiasStdDev", b"networkInitBiasStdDev", "networkInitWeightsStdDev", b"networkInitWeightsStdDev", "networkL1lambda", b"networkL1lambda", "networkMaskingFrac", b"networkMaskingFrac", "networkNumFeatures2", b"networkNumFeatures2", "niBaseNameTrain", b"niBaseNameTrain", "niPathTrain", b"niPathTrain", "radiusUm", b"radiusUm"]) -> None: ...

global___SpkSortTrainerDatasetParam = SpkSortTrainerDatasetParam

@typing.final
class SpkSortTrainerNetworkTrainingParams(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NUMEPOCHS_FIELD_NUMBER: builtins.int
    MINIBATCHSIZE_FIELD_NUMBER: builtins.int
    MAXNUMSAMPLES_FIELD_NUMBER: builtins.int
    ALPHA_FIELD_NUMBER: builtins.int
    ALPHABIAS_FIELD_NUMBER: builtins.int
    ERRORTOLPRETRAIN_FIELD_NUMBER: builtins.int
    NUMSAMPLESTESTMSE_FIELD_NUMBER: builtins.int
    numEpochs: builtins.int
    minibatchSize: builtins.int
    maxNumSamples: builtins.int
    alpha: builtins.float
    alphaBias: builtins.float
    errorTolPretrain: builtins.float
    numSamplesTestMSE: builtins.int
    def __init__(
        self,
        *,
        numEpochs: builtins.int | None = ...,
        minibatchSize: builtins.int | None = ...,
        maxNumSamples: builtins.int | None = ...,
        alpha: builtins.float | None = ...,
        alphaBias: builtins.float | None = ...,
        errorTolPretrain: builtins.float | None = ...,
        numSamplesTestMSE: builtins.int | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["alpha", b"alpha", "alphaBias", b"alphaBias", "errorTolPretrain", b"errorTolPretrain", "maxNumSamples", b"maxNumSamples", "minibatchSize", b"minibatchSize", "numEpochs", b"numEpochs", "numSamplesTestMSE", b"numSamplesTestMSE"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["alpha", b"alpha", "alphaBias", b"alphaBias", "errorTolPretrain", b"errorTolPretrain", "maxNumSamples", b"maxNumSamples", "minibatchSize", b"minibatchSize", "numEpochs", b"numEpochs", "numSamplesTestMSE", b"numSamplesTestMSE"]) -> None: ...

global___SpkSortTrainerNetworkTrainingParams = SpkSortTrainerNetworkTrainingParams

@typing.final
class SpkSortTrainerStoreDescriptor(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    BRAINREGION_FIELD_NUMBER: builtins.int
    NETWORKSIZE_FIELD_NUMBER: builtins.int
    NBRPATTERN_FIELD_NUMBER: builtins.int
    NETWORKID_FIELD_NUMBER: builtins.int
    NBHDDIM_FIELD_NUMBER: builtins.int
    RELEASE_FIELD_NUMBER: builtins.int
    brainRegion: builtins.str
    networkSize: global___SpkSortTrainerNetworkSize.ValueType
    nbrPattern: global___SpkSortTrainerNeighborPattern.ValueType
    networkID: builtins.str
    NbhdDim: builtins.int
    release: global___SpkSortTrainerNetworkRelease.ValueType
    def __init__(
        self,
        *,
        brainRegion: builtins.str | None = ...,
        networkSize: global___SpkSortTrainerNetworkSize.ValueType | None = ...,
        nbrPattern: global___SpkSortTrainerNeighborPattern.ValueType | None = ...,
        networkID: builtins.str | None = ...,
        NbhdDim: builtins.int | None = ...,
        release: global___SpkSortTrainerNetworkRelease.ValueType | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["NbhdDim", b"NbhdDim", "brainRegion", b"brainRegion", "nbrPattern", b"nbrPattern", "networkID", b"networkID", "networkSize", b"networkSize", "release", b"release"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["NbhdDim", b"NbhdDim", "brainRegion", b"brainRegion", "nbrPattern", b"nbrPattern", "networkID", b"networkID", "networkSize", b"networkSize", "release", b"release"]) -> None: ...

global___SpkSortTrainerStoreDescriptor = SpkSortTrainerStoreDescriptor

@typing.final
class SpkSortTrainerSetModeReq(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MODE_FIELD_NUMBER: builtins.int
    mode: builtins.int
    def __init__(
        self,
        *,
        mode: builtins.int | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["mode", b"mode"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["mode", b"mode"]) -> None: ...

global___SpkSortTrainerSetModeReq = SpkSortTrainerSetModeReq

@typing.final
class SpkSortTrainerCmdReq(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _SpkSortTrainerCmd:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _SpkSortTrainerCmdEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[SpkSortTrainerCmdReq._SpkSortTrainerCmd.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        CMD_ON: SpkSortTrainerCmdReq._SpkSortTrainerCmd.ValueType  # 0
        """must match DECsorterState enums"""
        CMD_OFF: SpkSortTrainerCmdReq._SpkSortTrainerCmd.ValueType  # 1

    class SpkSortTrainerCmd(_SpkSortTrainerCmd, metaclass=_SpkSortTrainerCmdEnumTypeWrapper): ...
    CMD_ON: SpkSortTrainerCmdReq.SpkSortTrainerCmd.ValueType  # 0
    """must match DECsorterState enums"""
    CMD_OFF: SpkSortTrainerCmdReq.SpkSortTrainerCmd.ValueType  # 1

    CMD_FIELD_NUMBER: builtins.int
    cmd: global___SpkSortTrainerCmdReq.SpkSortTrainerCmd.ValueType
    def __init__(
        self,
        *,
        cmd: global___SpkSortTrainerCmdReq.SpkSortTrainerCmd.ValueType | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["cmd", b"cmd"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["cmd", b"cmd"]) -> None: ...

global___SpkSortTrainerCmdReq = SpkSortTrainerCmdReq

@typing.final
class SpkSortTrainerStatusReply(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MSGJSON_FIELD_NUMBER: builtins.int
    msgJson: builtins.bytes
    def __init__(
        self,
        *,
        msgJson: builtins.bytes | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["msgJson", b"msgJson"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["msgJson", b"msgJson"]) -> None: ...

global___SpkSortTrainerStatusReply = SpkSortTrainerStatusReply

@typing.final
class WrangleRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MODE_FIELD_NUMBER: builtins.int
    SPEC_FIELD_NUMBER: builtins.int
    mode: global___WrangleMode.ValueType
    @property
    def spec(self) -> global___WrangleMerge: ...
    def __init__(
        self,
        *,
        mode: global___WrangleMode.ValueType | None = ...,
        spec: global___WrangleMerge | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["mode", b"mode", "spec", b"spec"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["mode", b"mode", "spec", b"spec"]) -> None: ...

global___WrangleRequest = WrangleRequest

@typing.final
class WrangleSpec(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    IMPORT_FIELD_NUMBER: builtins.int
    MERGE_FIELD_NUMBER: builtins.int
    @property
    def merge(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___WrangleMerge]: ...
    def __init__(
        self,
        *,
        merge: collections.abc.Iterable[global___WrangleMerge] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["import", b"import", "merge", b"merge"]) -> None: ...

global___WrangleSpec = WrangleSpec

@typing.final
class WrangleImport(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VEXSPIKESSIM_FIELD_NUMBER: builtins.int
    @property
    def vexSpikesSim(self) -> global___WrangleImportVexSpikesSim: ...
    def __init__(
        self,
        *,
        vexSpikesSim: global___WrangleImportVexSpikesSim | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["vexSpikesSim", b"vexSpikesSim"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["vexSpikesSim", b"vexSpikesSim"]) -> None: ...

global___WrangleImport = WrangleImport

@typing.final
class WrangleImportVexSpikesSim(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SOURCE_FIELD_NUMBER: builtins.int
    ISOVERWRITE_FIELD_NUMBER: builtins.int
    NOTE_FIELD_NUMBER: builtins.int
    isOverwrite: builtins.bool
    note: builtins.str
    @property
    def source(self) -> global___WrangleFileDesc: ...
    def __init__(
        self,
        *,
        source: global___WrangleFileDesc | None = ...,
        isOverwrite: builtins.bool | None = ...,
        note: builtins.str | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["isOverwrite", b"isOverwrite", "note", b"note", "source", b"source"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["isOverwrite", b"isOverwrite", "note", b"note", "source", b"source"]) -> None: ...

global___WrangleImportVexSpikesSim = WrangleImportVexSpikesSim

@typing.final
class WrangleFileDesc(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PATH_FIELD_NUMBER: builtins.int
    BASENAME_FIELD_NUMBER: builtins.int
    FILETYPE_FIELD_NUMBER: builtins.int
    path: builtins.str
    baseName: builtins.str
    fileType: builtins.str
    def __init__(
        self,
        *,
        path: builtins.str | None = ...,
        baseName: builtins.str | None = ...,
        fileType: builtins.str | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["baseName", b"baseName", "fileType", b"fileType", "path", b"path"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["baseName", b"baseName", "fileType", b"fileType", "path", b"path"]) -> None: ...

global___WrangleFileDesc = WrangleFileDesc

@typing.final
class WrangleMerge(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VEXSPIKESVIEW_FIELD_NUMBER: builtins.int
    @property
    def vexSpikesView(self) -> global___WrangleMergeVexSpikesView: ...
    def __init__(
        self,
        *,
        vexSpikesView: global___WrangleMergeVexSpikesView | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["vexSpikesView", b"vexSpikesView"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["vexSpikesView", b"vexSpikesView"]) -> None: ...

global___WrangleMerge = WrangleMerge

@typing.final
class WrangleMergeVexSpikesView(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SOURCEVEX_FIELD_NUMBER: builtins.int
    SOURCEBIO_FIELD_NUMBER: builtins.int
    VIEW_FIELD_NUMBER: builtins.int
    NOTE_FIELD_NUMBER: builtins.int
    ISBANDPASSFILTER_FIELD_NUMBER: builtins.int
    BANDPASSFILTERSPEC_FIELD_NUMBER: builtins.int
    ISOVERWRITE_FIELD_NUMBER: builtins.int
    SPIKEWINDOWMS_FIELD_NUMBER: builtins.int
    SPIKESHADOWMS_FIELD_NUMBER: builtins.int
    note: builtins.str
    isBandpassFilter: builtins.bool
    isOverwrite: builtins.bool
    spikeShadowMs: builtins.float
    @property
    def sourceVex(self) -> global___WrangleFileDesc: ...
    @property
    def sourceBio(self) -> global___WrangleFileDesc: ...
    @property
    def view(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___WrangleMergeVexSpikesViewSpec]: ...
    @property
    def bandpassFilterSpec(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.float]: ...
    @property
    def spikeWindowMs(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.float]: ...
    def __init__(
        self,
        *,
        sourceVex: global___WrangleFileDesc | None = ...,
        sourceBio: global___WrangleFileDesc | None = ...,
        view: collections.abc.Iterable[global___WrangleMergeVexSpikesViewSpec] | None = ...,
        note: builtins.str | None = ...,
        isBandpassFilter: builtins.bool | None = ...,
        bandpassFilterSpec: collections.abc.Iterable[builtins.float] | None = ...,
        isOverwrite: builtins.bool | None = ...,
        spikeWindowMs: collections.abc.Iterable[builtins.float] | None = ...,
        spikeShadowMs: builtins.float | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["isBandpassFilter", b"isBandpassFilter", "isOverwrite", b"isOverwrite", "note", b"note", "sourceBio", b"sourceBio", "sourceVex", b"sourceVex", "spikeShadowMs", b"spikeShadowMs"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["bandpassFilterSpec", b"bandpassFilterSpec", "isBandpassFilter", b"isBandpassFilter", "isOverwrite", b"isOverwrite", "note", b"note", "sourceBio", b"sourceBio", "sourceVex", b"sourceVex", "spikeShadowMs", b"spikeShadowMs", "spikeWindowMs", b"spikeWindowMs", "view", b"view"]) -> None: ...

global___WrangleMergeVexSpikesView = WrangleMergeVexSpikesView

@typing.final
class WrangleMergeVexSpikesViewSpec(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SINKBIO_FIELD_NUMBER: builtins.int
    SITENEIGHBORRADIUSUM_FIELD_NUMBER: builtins.int
    REQNUMSITENEIGHBORS_FIELD_NUMBER: builtins.int
    SPIKEWINDOWMS_FIELD_NUMBER: builtins.int
    SPIKESHADOWMS_FIELD_NUMBER: builtins.int
    NOTE_FIELD_NUMBER: builtins.int
    siteNeighborRadiusUm: builtins.float
    reqNumSiteNeighbors: builtins.int
    spikeShadowMs: builtins.float
    note: builtins.str
    @property
    def sinkBio(self) -> global___WrangleFileDesc: ...
    @property
    def spikeWindowMs(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.float]:
        """pre-peak, post-peak wrt reference time"""

    def __init__(
        self,
        *,
        sinkBio: global___WrangleFileDesc | None = ...,
        siteNeighborRadiusUm: builtins.float | None = ...,
        reqNumSiteNeighbors: builtins.int | None = ...,
        spikeWindowMs: collections.abc.Iterable[builtins.float] | None = ...,
        spikeShadowMs: builtins.float | None = ...,
        note: builtins.str | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["note", b"note", "reqNumSiteNeighbors", b"reqNumSiteNeighbors", "sinkBio", b"sinkBio", "siteNeighborRadiusUm", b"siteNeighborRadiusUm", "spikeShadowMs", b"spikeShadowMs"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["note", b"note", "reqNumSiteNeighbors", b"reqNumSiteNeighbors", "sinkBio", b"sinkBio", "siteNeighborRadiusUm", b"siteNeighborRadiusUm", "spikeShadowMs", b"spikeShadowMs", "spikeWindowMs", b"spikeWindowMs"]) -> None: ...

global___WrangleMergeVexSpikesViewSpec = WrangleMergeVexSpikesViewSpec

@typing.final
class RecgenImport(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FILEPATHS_FIELD_NUMBER: builtins.int
    OUTPATHS_FIELD_NUMBER: builtins.int
    ISIMPORTXDAT_FIELD_NUMBER: builtins.int
    ISIMPORTSPIKES_FIELD_NUMBER: builtins.int
    SCRAPENEIGHBORHOODUM_FIELD_NUMBER: builtins.int
    isImportXdat: builtins.bool
    isImportSpikes: builtins.bool
    scrapeNeighborhoodUm: builtins.float
    @property
    def filePaths(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    @property
    def outPaths(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    def __init__(
        self,
        *,
        filePaths: collections.abc.Iterable[builtins.str] | None = ...,
        outPaths: collections.abc.Iterable[builtins.str] | None = ...,
        isImportXdat: builtins.bool | None = ...,
        isImportSpikes: builtins.bool | None = ...,
        scrapeNeighborhoodUm: builtins.float | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["isImportSpikes", b"isImportSpikes", "isImportXdat", b"isImportXdat", "scrapeNeighborhoodUm", b"scrapeNeighborhoodUm"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["filePaths", b"filePaths", "isImportSpikes", b"isImportSpikes", "isImportXdat", b"isImportXdat", "outPaths", b"outPaths", "scrapeNeighborhoodUm", b"scrapeNeighborhoodUm"]) -> None: ...

global___RecgenImport = RecgenImport



"""Tests for FlexTimeRange and FlexSignalSpec coercion aliases."""

from __future__ import annotations

import pytest
from pydantic import BaseModel, ConfigDict, ValidationError

from radiens_core.models.dataset import FlexTimeRange, TimeRangeSpec, TrsMode
from radiens_core.models.signals import FlexSignalSpec, SignalSpec, SignalType

# ===================================================================
# Helper models that use the flex aliases
# ===================================================================


class _TimeRangeModel(BaseModel):
    model_config = ConfigDict(frozen=True)
    tr: FlexTimeRange


class _SignalSpecModel(BaseModel):
    model_config = ConfigDict(frozen=True)
    sig: FlexSignalSpec


# ===================================================================
# FlexTimeRange
# ===================================================================


class TestFlexTimeRange:
    def test_passthrough_spec(self) -> None:
        spec = TimeRangeSpec.subset(0, 10)
        m = _TimeRangeModel(tr=spec)
        assert m.tr is spec

    def test_list_subset(self) -> None:
        m = _TimeRangeModel(tr=[0, 10])
        assert isinstance(m.tr, TimeRangeSpec)
        assert m.tr.mode == TrsMode.SUBSET
        assert m.tr.start == 0
        assert m.tr.end == 10

    def test_tuple_subset(self) -> None:
        m = _TimeRangeModel(tr=(1.5, 3.5))
        assert isinstance(m.tr, TimeRangeSpec)
        assert m.tr.mode == TrsMode.SUBSET
        assert m.tr.start == 1.5
        assert m.tr.end == 3.5

    def test_list_lookback(self) -> None:

        m = _TimeRangeModel(tr=[5.0, float("nan")])
        assert isinstance(m.tr, TimeRangeSpec)
        assert m.tr.mode == TrsMode.FROM_HEAD
        assert m.tr.start == 5.0
        assert m.tr.end is None

    def test_invalid_type_raises(self) -> None:
        with pytest.raises(ValidationError):
            _TimeRangeModel(tr="invalid")  # type: ignore[arg-type]

    def test_invalid_type_string_raises(self) -> None:
        with pytest.raises(ValidationError):
            _TimeRangeModel(tr=42)  # type: ignore[arg-type]

    def test_wrong_length_list_raises(self) -> None:
        with pytest.raises(ValidationError):
            _TimeRangeModel(tr=[1, 2, 3])


# ===================================================================
# FlexSignalSpec
# ===================================================================


class TestFlexSignalSpec:
    def test_passthrough_spec(self) -> None:
        spec = SignalSpec.amp([0, 1, 2])
        m = _SignalSpecModel(sig=spec)
        assert m.sig is spec

    def test_list_of_ints_becomes_amp(self) -> None:
        m = _SignalSpecModel(sig=[0, 1, 2])
        assert isinstance(m.sig, SignalSpec)
        assert m.sig.sig_type == SignalType.AMP
        assert m.sig.channels == (0, 1, 2)

    def test_dict_becomes_multi(self) -> None:
        m = _SignalSpecModel(sig={"amp": [0, 1], "din": "all"})
        assert isinstance(m.sig, SignalSpec)
        assert m.sig.is_multi()
        assert m.sig.multi_select is not None
        assert SignalType.AMP in m.sig.multi_select
        assert SignalType.DIN in m.sig.multi_select

    def test_invalid_type_raises(self) -> None:
        with pytest.raises(ValidationError):
            _SignalSpecModel(sig="invalid")

    def test_invalid_type_int_raises(self) -> None:
        with pytest.raises(ValidationError):
            _SignalSpecModel(sig=42)  # type: ignore[arg-type]

    def test_empty_dict_raises(self) -> None:
        with pytest.raises(ValidationError):
            _SignalSpecModel(sig={})



"""Stimulation step control converters — domain ↔ proto."""

from __future__ import annotations

from radiens_core._generated import allegoserver_pb2
from radiens_core.models.stim import (
    StimStep,
    StimStepRequest,
    StimStepState,
)

# --- StimStep enum conversion ---

_STIM_STEP_TO_PROTO: dict[StimStep, allegoserver_pb2.StimStep.ValueType] = {
    StimStep.STIM_STEP_SIZE_MIN: allegoserver_pb2.StimStepSizeMin,
    StimStep.STIM_STEP_SIZE_10_NA: allegoserver_pb2.StimStepSize10nA,
    StimStep.STIM_STEP_SIZE_20_NA: allegoserver_pb2.StimStepSize20nA,
    StimStep.STIM_STEP_SIZE_50_NA: allegoserver_pb2.StimStepSize50nA,
    StimStep.STIM_STEP_SIZE_100_NA: allegoserver_pb2.StimStepSize100nA,
    StimStep.STIM_STEP_SIZE_200_NA: allegoserver_pb2.StimStepSize200nA,
    StimStep.STIM_STEP_SIZE_500_NA: allegoserver_pb2.StimStepSize500nA,
    StimStep.STIM_STEP_SIZE_1_UA: allegoserver_pb2.StimStepSize1uA,
    StimStep.STIM_STEP_SIZE_2_UA: allegoserver_pb2.StimStepSize2uA,
    StimStep.STIM_STEP_SIZE_5_UA: allegoserver_pb2.StimStepSize5uA,
    StimStep.STIM_STEP_SIZE_10_UA: allegoserver_pb2.StimStepSize10uA,
    StimStep.STIM_STEP_SIZE_MAX: allegoserver_pb2.StimStepSizeMax,
}

_STIM_STEP_FROM_PROTO: dict[int, StimStep] = {
    int(allegoserver_pb2.StimStepSizeMin): StimStep.STIM_STEP_SIZE_MIN,
    int(allegoserver_pb2.StimStepSize10nA): StimStep.STIM_STEP_SIZE_10_NA,
    int(allegoserver_pb2.StimStepSize20nA): StimStep.STIM_STEP_SIZE_20_NA,
    int(allegoserver_pb2.StimStepSize50nA): StimStep.STIM_STEP_SIZE_50_NA,
    int(allegoserver_pb2.StimStepSize100nA): StimStep.STIM_STEP_SIZE_100_NA,
    int(allegoserver_pb2.StimStepSize200nA): StimStep.STIM_STEP_SIZE_200_NA,
    int(allegoserver_pb2.StimStepSize500nA): StimStep.STIM_STEP_SIZE_500_NA,
    int(allegoserver_pb2.StimStepSize1uA): StimStep.STIM_STEP_SIZE_1_UA,
    int(allegoserver_pb2.StimStepSize2uA): StimStep.STIM_STEP_SIZE_2_UA,
    int(allegoserver_pb2.StimStepSize5uA): StimStep.STIM_STEP_SIZE_5_UA,
    int(allegoserver_pb2.StimStepSize10uA): StimStep.STIM_STEP_SIZE_10_UA,
    int(allegoserver_pb2.StimStepSizeMax): StimStep.STIM_STEP_SIZE_MAX,
}


def stim_step_to_proto(ss: StimStep) -> allegoserver_pb2.StimStep.ValueType:
    """Convert domain StimStep to proto StimStep."""
    return _STIM_STEP_TO_PROTO[ss]


def stim_step_from_proto(pb: int) -> StimStep:
    """Convert proto StimStep int to domain StimStep."""
    return _STIM_STEP_FROM_PROTO[pb]


def stim_step_request_to_proto(
    request: StimStepRequest,
) -> allegoserver_pb2.StimStepMessage:
    """Convert ``StimStepRequest`` domain model → proto message."""
    return allegoserver_pb2.StimStepMessage(
        stimStep=stim_step_to_proto(request.stim_step),
    )


def stim_step_state_from_proto(
    response: allegoserver_pb2.StimStepMessage,
) -> StimStepState:
    """Convert ``StimStepMessage`` proto response → domain model."""
    return StimStepState(
        stim_step=stim_step_from_proto(response.stimStep),
    )


def build_stim_step_proto(step: StimStep) -> allegoserver_pb2.StimStepMessage:
    """Build a StimStepMessage proto directly from a StimStep."""
    return allegoserver_pb2.StimStepMessage(
        stimStep=stim_step_to_proto(step),
    )



"""KPI (Key Performance Indicator) signal metrics domain models and enums."""

from __future__ import annotations

from enum import IntEnum

import numpy as np
import numpy.typing as npt
from pydantic import BaseModel, ConfigDict

from radiens_core.models.signals import SignalType
from radiens_core.models.time_range import TimeRange


class KpiMode(IntEnum):
    """KPI computation mode. Values match proto ``common.KpiMetricsMode``."""

    BASE = 0
    AVG = 1
    STREAM = 2


class KpiMetric(IntEnum):
    """KPI metric identifier. Values match proto ``common.KpiMetricsEnum``."""

    NUM_PTS = 0
    NUM_PTS_ISI = 1
    DUR_SEC = 2
    MEAN = 3
    MIN = 4
    MIN_ISI = 5
    MAX = 6
    MAX_ISI = 7
    MAX_ABS = 8
    MAX_ABS_ISI = 9
    TIMESTAMP_MIN = 10
    TIMESTAMP_MAX = 11
    MAX_MIN_DIFF_ABS = 12
    MAX_MIN_DIFF_ABS_ISI = 13
    SD = 14
    SD_ISI = 15
    VAR = 16
    VAR_ISI = 17
    RMS = 18
    RMS_ISI = 19
    NOISE_UV = 20
    SNR = 21
    NUM_EVENTS = 22
    EVENT_RATE = 23
    EVENT_MAX = 24
    EVENT_MIN = 25
    EVENT_MAX_ABS = 26
    EVENT_MAX_MIN_DIFF_ABS = 27
    EVENT_TIMESTAMP_MIN = 28
    EVENT_TIMESTAMP_MAX = 29
    EVENT_TIMESTAMP_MAX_ABS = 30
    EVENT_TIMESTAMP_MAX_MIN_DIFF_ABS = 31
    MEAN_MAX = 32
    MEAN_MIN = 33
    MEAN_MAX_ABS = 34
    EVENT_MEAN_MAX_MIN_DIFF_ABS = 35
    MAX_MIN_DIFF_ABS_AMPLIFIED = 36


class KpiMetricId(BaseModel):
    """Identifies a single KPI metric (mode + name pair).

    Attributes:
    mode
        The computation mode (BASE, AVG, or STREAM).
    name
        The metric to compute.
    """

    model_config = ConfigDict(frozen=True)

    mode: KpiMode
    name: KpiMetric


class KpiMetricsResult(BaseModel):
    """Result of a KPI metrics query.

    ``values`` is a 2-D numpy array of shape ``(n_channels, n_metrics)``
    where rows correspond to ``channel_indices`` and columns to ``metrics``.

    Attributes:
    channel_indices
        Native channel indices (ntvIdxs) in row order.
    metrics
        KPI metric IDs in column order.
    signal_type
        Signal type for all channels in this result.
    values
        Metric values: shape ``(n_channels, n_metrics)``.
    time_range
        Time range over which these metrics were computed.
    """

    model_config = ConfigDict(frozen=True, arbitrary_types_allowed=True)

    channel_indices: tuple[int, ...]
    metrics: tuple[KpiMetricId, ...]
    signal_type: SignalType
    values: npt.NDArray[np.float64]
    time_range: TimeRange


class KpiStatus(BaseModel):
    """Status of the KPI computation cache.

    Attributes:
    stream_group_id
        Stream group or datasource ID this status belongs to.
    num_packets_memory
        Number of KPI packets held in memory.
    packet_dur
        Duration of each KPI packet in seconds.
    update_period
        How often the KPI cache is updated, in seconds.
    persistence
        Exponential averaging persistence factor.
    beta
        Beta parameter for the averaging filter.
    is_tracking_signal_cache
        Whether the KPI cache is currently tracking the live signal cache.
    num_packets
        Total number of packets processed.
    memory_bytes
        Approximate memory used by the KPI cache in bytes.
    num_sigs
        Number of signals being tracked (``None`` if not yet active).
    time_range
        Time range covered by the current KPI cache (``None`` if not yet active).
    """

    model_config = ConfigDict(frozen=True)

    stream_group_id: str
    num_packets_memory: int
    packet_dur: float
    update_period: float
    persistence: float
    beta: float
    is_tracking_signal_cache: bool
    num_packets: int
    memory_bytes: float
    num_sigs: int | None = None
    time_range: TimeRange | None = None



"""Converters for core hardware configuration models."""

from radiens_core._converters._common_enums import (
    backbone_mode_from_proto,
    port_to_proto,
)
from radiens_core._generated import allegoserver_pb2
from radiens_core.models.core_config import CableDelay, CoreConfig, SetCoreConfigRequest
from radiens_core.models.ports import Port


def core_config_from_proto(proto: allegoserver_pb2.ConfigCore) -> CoreConfig:
    """Convert ConfigCore proto to domain model."""
    cable_delays = {}
    for key, val in proto.cableDelays.cableDelays.items():
        cable_delays[key] = CableDelay(
            port=Port(val.port),
            cable_delay=val.cableDelay,
            is_auto=val.isAuto,
        )

    return CoreConfig(
        allego_core_server_port=proto.allegoCoreServerPort,
        base_samp_freq=proto.baseSampFreq,
        stream_loop_dur_ms=proto.streamLoopDurMs,
        backbone_mode=backbone_mode_from_proto(proto.backboneMode),
        has_xdaq_gpio_expander=proto.hasXDAQGPIOExpander,
        cable_delays=cable_delays,
    )


def set_core_config_to_proto(
    req: SetCoreConfigRequest,
) -> allegoserver_pb2.SetConfigCoreRequest:
    """Convert SetCoreConfigRequest domain model to proto."""
    return build_update_core_config_proto(
        samp_freq=req.samp_freq,
        loop_dur=req.loop_dur,
        pcache_persistence=req.pcache_persistence,
        cable_delay=req.cable_delay,
    )


def build_update_core_config_proto(
    *,
    samp_freq: float | None = None,
    loop_dur: float | None = None,
    pcache_persistence: float | None = None,
    cable_delay: CableDelay | None = None,
) -> allegoserver_pb2.SetConfigCoreRequest:
    """Build a SetConfigCoreRequest proto from individual parameters."""
    proto = allegoserver_pb2.SetConfigCoreRequest()

    if samp_freq is not None:
        proto.sampFreq = samp_freq
    if loop_dur is not None:
        proto.loopDur = loop_dur
    if pcache_persistence is not None:
        proto.pcachePersistence = pcache_persistence
    if cable_delay is not None:
        proto.cableDelay.port = port_to_proto(cable_delay.port)
        proto.cableDelay.cableDelay = cable_delay.cable_delay
        proto.cableDelay.isAuto = cable_delay.is_auto

    return proto



"""Recording config converters — domain models <-> proto messages."""

from __future__ import annotations

from radiens_core._generated import allegoserver_pb2
from radiens_core.models.status import DataRecordingType, RecordingConfig

# --- DataRecordingType enum conversion ---

_DATA_RECORDING_TYPE_TO_PROTO: dict[DataRecordingType, allegoserver_pb2.DataRecordingType.ValueType] = {
    DataRecordingType.CONTINUOUS: allegoserver_pb2.continuous,
    DataRecordingType.SPIKES: allegoserver_pb2.spikes,
    DataRecordingType.CONTINUOUS_AND_SPIKES: allegoserver_pb2.continuousAndSpikes,
}

_DATA_RECORDING_TYPE_FROM_PROTO: dict[int, DataRecordingType] = {
    int(allegoserver_pb2.continuous): DataRecordingType.CONTINUOUS,
    int(allegoserver_pb2.spikes): DataRecordingType.SPIKES,
    int(allegoserver_pb2.continuousAndSpikes): DataRecordingType.CONTINUOUS_AND_SPIKES,
}


def data_recording_type_to_proto(drt: DataRecordingType) -> allegoserver_pb2.DataRecordingType.ValueType:
    """Convert domain DataRecordingType to proto DataRecordingType."""
    return _DATA_RECORDING_TYPE_TO_PROTO[drt]


def data_recording_type_from_proto(pb: int) -> DataRecordingType:
    """Convert proto DataRecordingType int to domain DataRecordingType."""
    return _DATA_RECORDING_TYPE_FROM_PROTO[pb]


def recording_config_to_proto(
    config: RecordingConfig,
) -> allegoserver_pb2.ConfigRecording:
    """Convert domain RecordingConfig to proto ConfigRecording."""
    return allegoserver_pb2.ConfigRecording(
        baseFileName=config.base_file_name,
        baseFilePath=config.base_file_path,
        dataSourceIdx=config.datasource_idx,
        timeStamp=config.time_stamp,
        splitFilesByPort=config.split_files_by_port,
        forceStartAtTStamp0=config.force_start_at_tstamp_0,
        recordingType=data_recording_type_to_proto(config.recording_type),
    )


def recording_config_from_proto(
    proto: allegoserver_pb2.ConfigRecording,
) -> RecordingConfig:
    """Convert proto ConfigRecording to domain RecordingConfig."""
    return RecordingConfig(
        base_file_name=proto.baseFileName,
        base_file_path=proto.baseFilePath,
        datasource_idx=proto.dataSourceIdx,
        time_stamp=proto.timeStamp,
        split_files_by_port=proto.splitFilesByPort,
        force_start_at_tstamp_0=proto.forceStartAtTStamp0,
        recording_type=data_recording_type_from_proto(proto.recordingType),
    )



"""Converter for DSP domain models to/from proto."""

from __future__ import annotations

from typing import cast

from radiens_core._generated import common_pb2
from radiens_core.models.dsp import DSPGroup, DSPParams, DSPType, FilterStage, FreqSpecBand
from radiens_core.models.ports import Port


def proto_to_domain_dsp_params(proto: common_pb2.DSPParams) -> DSPParams:
    """Convert DSPParams from proto to domain model."""

    freq_spec_band = None
    if proto.HasField("freqSpecBand"):
        freq_spec_band = FreqSpecBand(
            low_freq=proto.freqSpecBand.lowFreq,
            high_freq=proto.freqSpecBand.highFreq,
        )

    return DSPParams(
        type=DSPType(proto.type),
        stage=FilterStage(proto.stage),
        freq=proto.freq if proto.HasField("freq") else None,
        freq_spec_band=freq_spec_band,
        ref_ntv_chan_idx=proto.refNtvChanIdx if proto.HasField("refNtvChanIdx") else None,
        notch_bandwidth=proto.notchBandwidth if proto.HasField("notchBandwidth") else None,
        user_label=proto.userLabel,
        port=Port(proto.port) if proto.HasField("port") else None,
        target_ntv_chan_idx=proto.targetNtvChanIdx if proto.HasField("targetNtvChanIdx") else None,
        is_aux=proto.isAux,
        aux_chan_idx=proto.auxChanIdx if proto.HasField("auxChanIdx") else None,
        filter_order=proto.filterOrder if proto.HasField("filterOrder") else None,
        zero_phase=proto.zeroPhase,
        force_sos=proto.forceSOS,
    )


def domain_to_proto_dsp_params(params: DSPParams) -> common_pb2.DSPParams:
    """Convert DSPParams from domain model to proto."""

    proto = common_pb2.DSPParams(
        type=cast("common_pb2.DSPType.ValueType", params.type.value),
        stage=cast("common_pb2.FilterStage.ValueType", params.stage.value),
        userLabel=params.user_label,
        isAux=params.is_aux,
        zeroPhase=params.zero_phase,
        forceSOS=params.force_sos,
    )

    if params.freq is not None:
        proto.freq = params.freq

    if params.freq_spec_band is not None:
        proto.freqSpecBand.lowFreq = params.freq_spec_band.low_freq
        proto.freqSpecBand.highFreq = params.freq_spec_band.high_freq

    if params.ref_ntv_chan_idx is not None:
        proto.refNtvChanIdx = params.ref_ntv_chan_idx

    if params.notch_bandwidth is not None:
        proto.notchBandwidth = params.notch_bandwidth

    if params.port is not None:
        proto.port = cast("common_pb2.Port.ValueType", params.port.value)

    if params.target_ntv_chan_idx is not None:
        proto.targetNtvChanIdx = params.target_ntv_chan_idx

    if params.aux_chan_idx is not None:
        proto.auxChanIdx = params.aux_chan_idx

    if params.filter_order is not None:
        proto.filterOrder = params.filter_order

    return proto


def proto_to_domain_dsp_group(proto: common_pb2.DSPGroup) -> DSPGroup:
    """Convert DSPGroup from proto to domain model."""

    return DSPGroup(
        hardware=[proto_to_domain_dsp_params(p) for p in proto.hardware],
        software_stage1=[proto_to_domain_dsp_params(p) for p in proto.softwareStage1],
        software_stage2=[proto_to_domain_dsp_params(p) for p in proto.softwareStage2],
        spike_sorter=[proto_to_domain_dsp_params(p) for p in proto.spikeSorter],
    )



"""gRPC channel pooling with lifecycle management."""

from __future__ import annotations

import contextlib

import grpc

from radiens_core._constants import MAX_PB_MSG_SIZE_BYTES

_CHANNEL_OPTIONS: list[tuple[str, int | str | None]] = [
    ("grpc.max_receive_message_length", int(MAX_PB_MSG_SIZE_BYTES)),
    ("grpc.max_send_message_length", int(MAX_PB_MSG_SIZE_BYTES)),
]


class ChannelPool:
    """Manages gRPC channel lifecycle.

    Pools and reuses channels per address. Channels are created lazily
    and closed together on ``close()``.
    """

    def __init__(
        self,
        interceptors: list[grpc.UnaryUnaryClientInterceptor[object, object]] | None = None,
    ) -> None:
        self._channels: dict[str, grpc.Channel] = {}
        self._interceptors = interceptors or []
        self._closed = False

    def get_channel(self, address: str) -> grpc.Channel:
        """Get or create a gRPC channel for the given address."""
        if self._closed:
            raise RuntimeError("ChannelPool is closed")
        if address not in self._channels:
            base_channel = grpc.insecure_channel(address, options=_CHANNEL_OPTIONS)
            channel = grpc.intercept_channel(base_channel, *self._interceptors) if self._interceptors else base_channel
            self._channels[address] = channel
        return self._channels[address]

    def close(self) -> None:
        """Close all pooled channels."""
        if self._closed:
            return
        self._closed = True
        for channel in self._channels.values():
            with contextlib.suppress(Exception):
                channel.close()
        self._channels.clear()

    @property
    def is_closed(self) -> bool:
        return self._closed

    def __enter__(self) -> ChannelPool:
        return self

    def __exit__(self, *args: object) -> None:
        self.close()



"""Signal data <-> proto converters."""

from __future__ import annotations

import numpy as np
import numpy.typing as npt

from radiens_core._generated import common_pb2
from radiens_core.models.signals import SignalArrays, SigSelect


def raw_signals_to_numpy(raw: common_pb2.RawSignals) -> npt.NDArray[np.float64] | None:
    """Convert proto RawSignals to a numpy 2D array (channels x samples).

    Returns None if data is empty.
    """
    if len(raw.data) == 0:
        return None
    arr = np.frombuffer(raw.data, dtype=np.float32).astype(np.float64)
    return arr.reshape(int(raw.shape[0]), int(raw.shape[1]))


def signal_arrays_from_hd_snapshot(response: common_pb2.HDSnapshot) -> SignalArrays:
    """Build domain SignalArrays from proto HDSnapshot (old Pcache1/GetHDSnapshot API)."""
    return SignalArrays(
        amp=raw_signals_to_numpy(response.priSignals) if response.HasField("priSignals") else None,
        gpio_ain=raw_signals_to_numpy(response.auxSignals) if response.HasField("auxSignals") else None,
        gpio_din=raw_signals_to_numpy(response.dinSignals) if response.HasField("dinSignals") else None,
        gpio_dout=raw_signals_to_numpy(response.doutSignals) if response.HasField("doutSignals") else None,
    )


def radix_matrix_to_numpy(
    mb: common_pb2.RadixMatrixBytes,
) -> npt.NDArray[np.float64] | None:
    """Convert proto RadixMatrixBytes to a numpy 2D array (channels x samples).

    Returns None if data is empty.
    """
    if len(mb.data) == 0:
        return None
    arr = np.frombuffer(mb.data, dtype=np.float32).astype(np.float64)
    return arr.reshape(mb.shape[0], mb.shape[1])


def signal_arrays_from_signals2(
    sigs: common_pb2.Signals2,
) -> SignalArrays:
    """Build domain SignalArrays from proto Signals2."""
    return SignalArrays(
        amp=radix_matrix_to_numpy(sigs.amp),
        gpio_ain=radix_matrix_to_numpy(sigs.ain),
        gpio_din=radix_matrix_to_numpy(sigs.din),
        gpio_dout=radix_matrix_to_numpy(sigs.dout),
    )


def sig_select_to_proto(sig_sel: SigSelect) -> common_pb2.SigSelector:
    """Convert domain SigSelect to proto SigSelector."""
    return common_pb2.SigSelector(
        ampNtvIdxs=sig_sel.amp.tolist(),
        ainNtvIdxs=sig_sel.gpio_ain.tolist(),
        dinNtvIdxs=sig_sel.gpio_din.tolist(),
        doutNtvIdxs=sig_sel.gpio_dout.tolist(),
    )



"""Tests for channel metadata converters."""

from __future__ import annotations

from radiens_core._converters._channels import channel_metadata_from_proto
from radiens_core._generated import common_pb2


class TestChannelCoordinateScaling:
    def test_site_ctr_x_scaled_by_10(self) -> None:
        """siteCtrX in proto (10s of microns) must be multiplied by 10 in domain."""
        ch = common_pb2.ChannelRecord(
            chanType=common_pb2.PRI,
            siteCtrX=5.0,
            siteCtrY=0.0,
            siteCtrZ=0.0,
        )
        sig_group = common_pb2.SignalGroup()
        sig_group.channels.append(ch)

        metadata = channel_metadata_from_proto(sig_group)

        assert len(metadata.site_positions) == 1
        assert metadata.site_positions[0].probe_x == 50.0

    def test_all_site_coordinates_scaled(self) -> None:
        """All site coordinate fields must be scaled by 10."""
        ch = common_pb2.ChannelRecord(
            chanType=common_pb2.PRI,
            siteCtrX=1.0,
            siteCtrY=2.0,
            siteCtrZ=3.0,
        )
        sig_group = common_pb2.SignalGroup()
        sig_group.channels.append(ch)

        metadata = channel_metadata_from_proto(sig_group)

        pos = metadata.site_positions[0]
        assert pos.probe_x == 10.0
        assert pos.probe_y == 20.0
        assert pos.probe_z == 30.0



"""AllegoCore RPC wrappers — thin service layer (proto only)."""

from __future__ import annotations

from typing import TYPE_CHECKING

import grpc

from radiens_core._constants import PRIMARY_CACHE_STREAM_GROUP_ID
from radiens_core._generated import allegoserver_pb2, allegoserver_pb2_grpc, common_pb2, datasource_pb2
from radiens_core._transport._errors import handle_grpc_error
from radiens_core.models.common import ClientType, ServiceEndpoint

if TYPE_CHECKING:
    from radiens_core._transport._channel_pool import ChannelPool


def get_intan_impedance_proto(
    pool: ChannelPool,
    addr: str,
    request: datasource_pb2.ImpedanceRequest,
) -> datasource_pb2.Impedance:
    """Send ``GetIntanImpedance`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    try:
        return stub.GetIntanImpedance(request)
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise  # unreachable, but satisfies mypy


def scan_ports_proto(
    pool: ChannelPool,
    addr: str,
) -> common_pb2.SignalGroup:
    """Send ``ScanPorts`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    try:
        return stub.ScanPorts(common_pb2.StandardRequest())
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise  # unreachable, but satisfies mypy


def get_stim_params_proto(
    pool: ChannelPool,
    addr: str,
) -> allegoserver_pb2.StimParamsReply:
    """Send ``GetStimParams`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    try:
        return stub.GetStimParams(common_pb2.StandardRequest())
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise  # unreachable, but satisfies mypy


def get_status_proto(
    pool: ChannelPool,
    addr: str,
) -> allegoserver_pb2.BackboneStatus:
    """Send ``GetStatus`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    try:
        return stub.GetStatus(common_pb2.StandardRequest())
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise  # unreachable, but satisfies mypy


def set_stream_state_proto(
    pool: ChannelPool,
    addr: str,
    request: allegoserver_pb2.SetStreamRequest,
) -> common_pb2.StandardReply:
    """Send ``SetStreamState`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    try:
        return stub.SetStreamState(request)
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise  # unreachable, but satisfies mypy


def healthcheck_proto(
    pool: ChannelPool,
    addr: str,
    service: ServiceEndpoint = ServiceEndpoint.ALLEGO_LIFECYCLE,
) -> common_pb2.StandardReply:
    """Send ``Healthcheck`` RPC to an allego service, return raw proto response.

    *service* selects the stub: ``ALLEGO_CORE`` uses ``AllegoCoreStub``,
    anything else defaults to ``AllegoLifecycleStub``.
    """
    channel = pool.get_channel(addr)
    req = common_pb2.StandardRequest()
    try:
        if service == ServiceEndpoint.ALLEGO_CORE:
            return allegoserver_pb2_grpc.AllegoCoreStub(channel).Healthcheck(req)
        return allegoserver_pb2_grpc.AllegoLifecycleStub(channel).Healthcheck(req)
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise  # unreachable, but satisfies mypy


def set_record_state_proto(
    pool: ChannelPool,
    addr: str,
    request: allegoserver_pb2.SetRecordRequest,
) -> common_pb2.StandardReply:
    """Send ``SetRecordState`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    try:
        return stub.SetRecordState(request)
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise  # unreachable, but satisfies mypy


def restart_proto(
    pool: ChannelPool,
    addr: str,
    request: allegoserver_pb2.RestartRequest,
) -> common_pb2.StandardReply:
    """Send ``Restart`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    try:
        return stub.Restart(request)
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise  # unreachable, but satisfies mypy


def set_stim_params_proto(
    pool: ChannelPool,
    addr: str,
    request: allegoserver_pb2.StimParams,
) -> common_pb2.StandardReply:
    """Send ``SetStimParams`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    try:
        return stub.SetStimParams(request)
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise  # unreachable, but satisfies mypy


def set_trigger_state_proto(
    pool: ChannelPool,
    addr: str,
    request: allegoserver_pb2.SetTriggerRequest,
) -> common_pb2.StandardReply:
    """Send ``SetTriggerState`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    try:
        return stub.SetTriggerState(request)
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise  # unreachable, but satisfies mypy


def get_trigger_state_proto(
    pool: ChannelPool,
    addr: str,
) -> allegoserver_pb2.TriggerState:
    """Send ``GetTriggerState`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    try:
        return stub.GetTriggerState(common_pb2.StandardRequest())
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise  # unreachable, but satisfies mypy


def manual_stim_trigger_proto(
    pool: ChannelPool,
    addr: str,
    request: allegoserver_pb2.ManualStimTriggerRequest,
) -> common_pb2.StandardReply:
    """Send ``ManualStimTrigger`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    try:
        return stub.ManualStimTrigger(request)
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise  # unreachable, but satisfies mypy


def manual_stim_trigger_toggle_proto(
    pool: ChannelPool,
    addr: str,
    request: allegoserver_pb2.ManualStimTriggerToggleRequest,
) -> common_pb2.StandardReply:
    """Send ``ManualStimTriggerToggle`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    try:
        return stub.ManualStimTriggerToggle(request)
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise  # unreachable, but satisfies mypy


def set_stim_step_proto(
    pool: ChannelPool,
    addr: str,
    request: allegoserver_pb2.StimStepMessage,
) -> common_pb2.StandardReply:
    """Send ``SetStimStep`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    try:
        return stub.SetStimStep(request)
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise  # unreachable, but satisfies mypy


def get_stim_step_proto(
    pool: ChannelPool,
    addr: str,
) -> allegoserver_pb2.StimStepMessage:
    """Send ``GetStimStep`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    try:
        return stub.GetStimStep(common_pb2.StandardRequest())
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise  # unreachable, but satisfies mypy


def get_signal_group_proto(
    pool: ChannelPool,
    addr: str,
) -> common_pb2.SignalGroup:
    """Send ``GetSignalGroup`` RPC, return raw proto response."""
    req = common_pb2.SignalGroupIDRequest()
    # Use default stream group ID if none provided, like the old client
    req.streamGroupId = PRIMARY_CACHE_STREAM_GROUP_ID

    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    try:
        return stub.GetSignalGroup(req)
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise  # unreachable, but satisfies mypy


def set_recording_config_proto(
    pool: ChannelPool,
    addr: str,
    request: allegoserver_pb2.ConfigRecording,
) -> common_pb2.StandardReply:
    """Send ``SetConfigRecording`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    try:
        return stub.SetConfigRecording(request)
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise  # unreachable, but satisfies mypy


def get_recording_config_proto(
    pool: ChannelPool,
    addr: str,
) -> allegoserver_pb2.ConfigRecording:
    """Send ``GetConfigRecording`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    try:
        return stub.GetConfigRecording(common_pb2.StandardRequest())
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise  # unreachable, but satisfies mypy


def get_config_proto(
    pool: ChannelPool,
    addr: str,
) -> allegoserver_pb2.ConfigCore:
    """Send ``GetConfig`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    try:
        return stub.GetConfig(common_pb2.StandardRequest())
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise  # unreachable, but satisfies mypy


def set_config_core_proto(
    pool: ChannelPool,
    addr: str,
    request: allegoserver_pb2.SetConfigCoreRequest,
) -> common_pb2.StandardReply:
    """Send ``SetConfigCore`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    try:
        return stub.SetConfigCore(request)
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise  # unreachable, but satisfies mypy


def set_dac_gain_proto(
    pool: ChannelPool,
    addr: str,
    request: allegoserver_pb2.DACGainRequest,
) -> common_pb2.StandardReply:
    """Send ``SetDACGain`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    try:
        return stub.SetDACGain(request)
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise  # unreachable, but satisfies mypy


def set_dac_stream_proto(
    pool: ChannelPool,
    addr: str,
    request: common_pb2.DACStreamRequest,
) -> common_pb2.StandardReply:
    """Send ``SetDACStream`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    try:
        return stub.SetDACStream(request)
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise  # unreachable, but satisfies mypy


def set_dac_off_proto(
    pool: ChannelPool,
    addr: str,
    request: common_pb2.DACOffRequest,
) -> common_pb2.StandardReply:
    """Send ``SetDACOff`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    try:
        return stub.SetDACOff(request)
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise  # unreachable, but satisfies mypy


def set_dac_highpass_proto(
    pool: ChannelPool,
    addr: str,
    request: allegoserver_pb2.DACHighPassRegister,
) -> common_pb2.StandardReply:
    """Send ``SetDACHighPass`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    try:
        return stub.SetDACHighPass(request)
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise


# --- Digital Output (DIO) ---


def set_dio_events_proto(
    pool: ChannelPool,
    addr: str,
    request: allegoserver_pb2.DIOModeEventsRequest,
) -> common_pb2.StandardReply:
    """Send ``SetDIOEvents`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    try:
        return stub.SetDIOEvents(request)
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise


def set_dio_manual_proto(
    pool: ChannelPool,
    addr: str,
    request: allegoserver_pb2.DIOModeManualRequest,
) -> common_pb2.StandardReply:
    """Send ``SetDIOManual`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    try:
        return stub.SetDIOManual(request)
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise


def set_dio_gated_proto(
    pool: ChannelPool,
    addr: str,
    request: allegoserver_pb2.DIOModeGatedRequest,
) -> common_pb2.StandardReply:
    """Send ``SetDIOGated`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    try:
        return stub.SetDIOGated(request)
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise


def set_dio_pulse_proto(
    pool: ChannelPool,
    addr: str,
    request: allegoserver_pb2.DIOModePulseRequest,
) -> common_pb2.StandardReply:
    """Send ``SetDIOPulse`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    try:
        return stub.SetDIOPulse(request)
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise


def get_dio_reg_proto(
    pool: ChannelPool,
    addr: str,
) -> allegoserver_pb2.DigitalOutRegister:
    """Send ``GetDIOReg`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    try:
        return stub.GetDIOReg(common_pb2.StandardRequest())
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise


# unreachable, but satisfies mypy


def get_dac_reg_proto(
    pool: ChannelPool,
    addr: str,
) -> allegoserver_pb2.AnalogOutRegister:
    """Send ``GetDACReg`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    try:
        return stub.GetDACReg(common_pb2.StandardRequest())
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise  # unreachable, but satisfies mypy


def get_dsp_group_proto(
    pool: ChannelPool,
    addr: str,
    stream_group_id: str,
) -> common_pb2.DSPGroup:
    """Send ``GetDSPGroup`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    req = common_pb2.SignalGroupIDRequest(streamGroupId=stream_group_id)
    try:
        return stub.GetDSPGroup(req)
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise


def set_dsp_group_proto(
    pool: ChannelPool,
    addr: str,
    request: common_pb2.SetDSPGroupRequest,
) -> common_pb2.StandardReply:
    """Send ``SetDSPGroup`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    try:
        return stub.SetDSPGroup(request)
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise


# --- Sinaps ---


def load_all_mosi_proto(
    pool: ChannelPool,
    addr: str,
    request: allegoserver_pb2.LoadAllMosiRequest,
) -> common_pb2.StandardReply:
    """Send ``LoadAllMosi`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    try:
        return stub.LoadAllMosi(request)
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise  # unreachable, but satisfies mypy


def transmit_mosi_proto(
    pool: ChannelPool,
    addr: str,
) -> common_pb2.StandardReply:
    """Send ``TransmitMosi`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    try:
        return stub.TransmitMosi(allegoserver_pb2.TransmitMosiRequest())
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise  # unreachable, but satisfies mypy


def read_wire_out_proto(
    pool: ChannelPool,
    addr: str,
    request: allegoserver_pb2.ReadWireOutRequest,
) -> allegoserver_pb2.ReadWireOutReply:
    """Send ``ReadWireOut`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    try:
        return stub.ReadWireOut(request)
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise  # unreachable, but satisfies mypy


def get_sinaps_status_registers_proto(
    pool: ChannelPool,
    addr: str,
) -> allegoserver_pb2.SinapsStatusRegisters:
    """Send ``GetSinapsStatusRegisters`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    try:
        return stub.GetSinapsStatusRegisters(common_pb2.StandardRequest())
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise  # unreachable, but satisfies mypy


def flash_sinaps_proto(
    pool: ChannelPool,
    addr: str,
    request: allegoserver_pb2.FlashSinapsRequest,
) -> common_pb2.StandardReply:
    """Send ``FlashSinaps`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = allegoserver_pb2_grpc.AllegoCoreStub(channel)
    try:
        return stub.FlashSinaps(request)
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.ALLEGO)
        raise  # unreachable, but satisfies mypy



"""Tests for CurateClient."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from collections.abc import Generator
from unittest.mock import MagicMock, patch

import pytest

from radiens_core._generated import common_pb2
from radiens_core.curate_client import CurateClient
from radiens_core.models.dataset import (
    CARParams,
    HighLowPassParams,
    ProtocolSpec,
    RadiensFileType,
    SinkParams,
    SourceParams,
    TransformEdge,
    TransformNode,
    TransformNodeType,
)


@pytest.fixture
def mock_discovery() -> Generator[Any, None, None]:  # type: ignore[explicit-any]
    """Mock server discovery."""
    with patch("radiens_core._base_file_client.discover_radiens_addresses") as mock:
        mock.return_value = {
            "RADIENS_CORE": "localhost:50055",
            "RADIENS_LIFECYCLE": "localhost:50055",
        }
        yield mock


@pytest.fixture
def mock_auth() -> Generator[Any, None, None]:  # type: ignore[explicit-any]
    """Mock authentication interceptor."""
    with patch("radiens_core._base_client.AuthInterceptor") as mock:
        yield mock


def test_execute_protocol_progress(mock_discovery: Any, mock_auth: Any) -> None:  # type: ignore[explicit-any]
    """Test _execute_protocol with a mocked progress stream."""
    client = CurateClient()

    # Mock set_protocol_proto
    mock_set = MagicMock()

    # Mock apply_protocol_proto to return a stream of progress messages
    mock_apply = MagicMock()

    def progress_stream(*args: Any, **kwargs: Any) -> Generator[common_pb2.ApplyProtocolProgress, None, None]:  # type: ignore[explicit-any]
        yield common_pb2.ApplyProtocolProgress(fracComplete=0.2)
        yield common_pb2.ApplyProtocolProgress(fracComplete=0.5)
        yield common_pb2.ApplyProtocolProgress(fracComplete=1.0)

    mock_apply.return_value = progress_stream()

    with (
        patch("radiens_core.curate_client.set_protocol_proto", mock_set),
        patch("radiens_core.curate_client.apply_protocol_proto", mock_apply),
        patch("radiens_core.curate_client.tqdm") as mock_tqdm,
    ):
        # Mock tqdm instance
        pbar = mock_tqdm.return_value.__enter__.return_value

        nodes = [common_pb2.TransformNode(id="node1")]
        edges = [common_pb2.TransformEdge(id="edge1")]

        client._execute_protocol(nodes, edges, desc="Test Task")

        # Verify set_protocol was called
        mock_set.assert_called_once()
        protocol = mock_set.call_args[0][2]
        assert protocol.transformNodes[0].id == "node1"
        assert protocol.transformEdges[0].id == "edge1"

        # Verify apply_protocol was called
        mock_apply.assert_called_once()

        # Verify tqdm updates
        # Expected updates: 0.2*100=20, (0.5-0.2)*100=30, (1.0-0.5)*100=50
        assert pbar.update.call_count == 3
        calls = pbar.update.call_args_list
        assert calls[0][0][0] == pytest.approx(20.0)
        assert calls[1][0][0] == pytest.approx(30.0)
        assert calls[2][0][0] == pytest.approx(50.0)


def test_link_data_file(mock_discovery: Any, mock_auth: Any) -> None:  # type: ignore[explicit-any]
    """Test link_data_file method calls the correct service."""
    client = CurateClient()
    mock_meta = MagicMock()
    mock_meta.id = "test_ds"

    with patch("radiens_core._base_file_client.set_datasource", return_value=mock_meta) as mock_set_ds:
        meta = client.link_data_file("/path/to/test.xdat")

        assert meta == mock_meta
        assert client._datasets["test_ds"] == mock_meta
        mock_set_ds.assert_called_once()
        assert mock_set_ds.call_args[0][2] == "/path/to/test.xdat"


def test_slice_time(mock_discovery: Any, mock_auth: Any) -> None:  # type: ignore[explicit-any]
    """Test slice_time method builds correct graph and executes."""
    client = CurateClient()

    mock_input_meta = MagicMock()
    mock_input_meta.id = "input_ds"
    mock_input_meta.base_name = "test_recording"
    mock_input_meta.path = "/path/to"
    mock_input_meta.file_type = RadiensFileType.XDAT
    client._datasets["input_ds"] = mock_input_meta

    mock_output_meta = MagicMock()

    with (
        patch.object(client, "_execute_protocol") as mock_exec,
        patch.object(client, "link_data_file", return_value=mock_output_meta),
    ):
        res = client.slice_time("input_ds", 10.5, 20.5, "output.xdat")

        assert res == mock_output_meta
        mock_exec.assert_called_once()
        args, kwargs = mock_exec.call_args
        nodes, edges = args
        assert "Slicing output.xdat" in kwargs["desc"]

        # Verify nodes: Source, Transform, Sink
        assert len(nodes) == 3
        assert nodes[0].type == common_pb2.SOURCE
        assert nodes[0].datasourceSinkParams.id == "input_ds"

        assert nodes[1].type == common_pb2.SLICE_TIME
        assert nodes[1].sliceTimeParams.timeStart == 10.5
        assert nodes[1].sliceTimeParams.timeEnd == 20.5

        assert nodes[2].type == common_pb2.SINK
        assert nodes[2].datasourceSinkParams.dsrcName == "output"

        # Verify edges
        assert len(edges) == 2
        assert edges[0].source == nodes[0].id
        assert edges[0].target == nodes[1].id
        assert edges[1].source == nodes[1].id
        assert edges[1].target == nodes[2].id


def test_downsample(mock_discovery: Any, mock_auth: Any) -> None:  # type: ignore[explicit-any]
    """Test downsample method builds correct graph and executes."""
    client = CurateClient()

    mock_input_meta = MagicMock()
    mock_input_meta.id = "input_ds"
    mock_input_meta.base_name = "test_recording"
    mock_input_meta.path = "/path/to"
    mock_input_meta.file_type = RadiensFileType.XDAT
    client._datasets["input_ds"] = mock_input_meta

    mock_output_meta = MagicMock()

    with (
        patch.object(client, "_execute_protocol") as mock_exec,
        patch.object(client, "link_data_file", return_value=mock_output_meta),
    ):
        res = client.downsample("input_ds", 4, "downsampled.rhd")

        assert res == mock_output_meta
        mock_exec.assert_called_once()
        args, kwargs = mock_exec.call_args
        nodes, _edges = args
        assert "Downsampling downsampled.rhd" in kwargs["desc"]

        assert len(nodes) == 3
        assert nodes[1].type == common_pb2.DOWNSAMPLE
        assert nodes[1].downsampleParams.sampleFactor == 4

        assert nodes[2].datasourceSinkParams.fileType == common_pb2.RHD


def _make_client_with_dataset(  # type: ignore[explicit-any]
    mock_discovery: Any,
    mock_auth: Any,
) -> tuple[CurateClient, MagicMock]:
    """Helper: create a CurateClient with a pre-linked mock dataset."""
    client = CurateClient()
    mock_input_meta = MagicMock()
    mock_input_meta.id = "input_ds"
    mock_input_meta.base_name = "test_recording"
    mock_input_meta.path = "/path/to"
    mock_input_meta.file_type = RadiensFileType.XDAT
    client._datasets["input_ds"] = mock_input_meta
    return client, mock_input_meta


def test_highpass(mock_discovery: Any, mock_auth: Any) -> None:  # type: ignore[explicit-any]
    """Test highpass builds correct TransformNode."""
    client, _ = _make_client_with_dataset(mock_discovery, mock_auth)
    mock_output = MagicMock()

    with (
        patch.object(client, "_execute_protocol") as mock_exec,
        patch.object(client, "link_data_file", return_value=mock_output),
    ):
        res = client.highpass("input_ds", 300.0, "hp.xdat")

        assert res == mock_output
        mock_exec.assert_called_once()
        nodes, _edges = mock_exec.call_args[0]

        assert len(nodes) == 3
        assert nodes[1].type == common_pb2.HIGHPASS_FILTER
        assert nodes[1].highLowPassParams.frequency == 300.0
        assert not nodes[1].highLowPassParams.HasField("order")


def test_highpass_with_order(mock_discovery: Any, mock_auth: Any) -> None:  # type: ignore[explicit-any]
    """Test highpass passes filter order when specified."""
    client, _ = _make_client_with_dataset(mock_discovery, mock_auth)
    mock_output = MagicMock()

    with (
        patch.object(client, "_execute_protocol"),
        patch.object(client, "link_data_file", return_value=mock_output),
    ):
        client.highpass("input_ds", 300.0, "hp.xdat", order=4)

        nodes, _edges = client._execute_protocol.call_args[0]  # type: ignore[attr-defined]
        assert nodes[1].highLowPassParams.order == 4


def test_lowpass(mock_discovery: Any, mock_auth: Any) -> None:  # type: ignore[explicit-any]
    """Test lowpass builds correct TransformNode."""
    client, _ = _make_client_with_dataset(mock_discovery, mock_auth)
    mock_output = MagicMock()

    with (
        patch.object(client, "_execute_protocol") as mock_exec,
        patch.object(client, "link_data_file", return_value=mock_output),
    ):
        res = client.lowpass("input_ds", 6000.0, "lp.xdat")

        assert res == mock_output
        nodes, _edges = mock_exec.call_args[0]

        assert nodes[1].type == common_pb2.LOWPASS_FILTER
        assert nodes[1].highLowPassParams.frequency == 6000.0


def test_bandpass(mock_discovery: Any, mock_auth: Any) -> None:  # type: ignore[explicit-any]
    """Test bandpass builds correct TransformNode."""
    client, _ = _make_client_with_dataset(mock_discovery, mock_auth)
    mock_output = MagicMock()

    with (
        patch.object(client, "_execute_protocol") as mock_exec,
        patch.object(client, "link_data_file", return_value=mock_output),
    ):
        res = client.bandpass("input_ds", 300.0, 6000.0, "bp.xdat")

        assert res == mock_output
        nodes, _edges = mock_exec.call_args[0]

        assert nodes[1].type == common_pb2.BANDPASS_FILTER
        assert nodes[1].bandParams.lowFrequency == 300.0
        assert nodes[1].bandParams.highFrequency == 6000.0
        assert not nodes[1].bandParams.HasField("order")


def test_bandpass_with_order(mock_discovery: Any, mock_auth: Any) -> None:  # type: ignore[explicit-any]
    """Test bandpass passes filter order when specified."""
    client, _ = _make_client_with_dataset(mock_discovery, mock_auth)

    with (
        patch.object(client, "_execute_protocol"),
        patch.object(client, "link_data_file", return_value=MagicMock()),
    ):
        client.bandpass("input_ds", 300.0, 6000.0, "bp.xdat", order=6)

        nodes, _edges = client._execute_protocol.call_args[0]  # type: ignore[attr-defined]
        assert nodes[1].bandParams.order == 6


def test_bandstop(mock_discovery: Any, mock_auth: Any) -> None:  # type: ignore[explicit-any]
    """Test bandstop builds correct TransformNode."""
    client, _ = _make_client_with_dataset(mock_discovery, mock_auth)
    mock_output = MagicMock()

    with (
        patch.object(client, "_execute_protocol") as mock_exec,
        patch.object(client, "link_data_file", return_value=mock_output),
    ):
        res = client.bandstop("input_ds", 58.0, 62.0, "bs.xdat")

        assert res == mock_output
        nodes, _edges = mock_exec.call_args[0]

        assert nodes[1].type == common_pb2.BANDSTOP_FILTER
        assert nodes[1].bandParams.lowFrequency == 58.0
        assert nodes[1].bandParams.highFrequency == 62.0


def test_notch(mock_discovery: Any, mock_auth: Any) -> None:  # type: ignore[explicit-any]
    """Test notch builds correct TransformNode."""
    client, _ = _make_client_with_dataset(mock_discovery, mock_auth)
    mock_output = MagicMock()

    with (
        patch.object(client, "_execute_protocol") as mock_exec,
        patch.object(client, "link_data_file", return_value=mock_output),
    ):
        res = client.notch("input_ds", 60.0, 4.0, "notched.xdat")

        assert res == mock_output
        nodes, _edges = mock_exec.call_args[0]

        assert nodes[1].type == common_pb2.NOTCH_FILTER
        assert nodes[1].notchParams.notchFrequency == 60.0
        assert nodes[1].notchParams.notchBandwidth == 4.0
        assert not nodes[1].notchParams.HasField("order")


def test_slice_channels(mock_discovery: Any, mock_auth: Any) -> None:  # type: ignore[explicit-any]
    """Test slice_channels builds correct TransformNode."""
    client, _ = _make_client_with_dataset(mock_discovery, mock_auth)
    mock_output = MagicMock()

    with (
        patch.object(client, "_execute_protocol") as mock_exec,
        patch.object(client, "link_data_file", return_value=mock_output),
    ):
        res = client.slice_channels("input_ds", [0, 2, 5, 10], "sliced_ch.xdat")

        assert res == mock_output
        nodes, _edges = mock_exec.call_args[0]

        assert len(nodes) == 3
        assert nodes[1].type == common_pb2.SLICE_CHANNELS
        assert list(nodes[1].sliceChannelsParams.sysChanIdxs) == [0, 2, 5, 10]


def test_slice_channels_graph_structure(mock_discovery: Any, mock_auth: Any) -> None:  # type: ignore[explicit-any]
    """Test slice_channels produces correct Source -> Transform -> Sink graph."""
    client, _ = _make_client_with_dataset(mock_discovery, mock_auth)

    with (
        patch.object(client, "_execute_protocol") as mock_exec,
        patch.object(client, "link_data_file", return_value=MagicMock()),
    ):
        client.slice_channels("input_ds", [0, 1], "out.xdat")

        nodes, edges = mock_exec.call_args[0]

        assert nodes[0].type == common_pb2.SOURCE
        assert nodes[0].datasourceSinkParams.id == "input_ds"
        assert nodes[2].type == common_pb2.SINK
        assert nodes[2].datasourceSinkParams.dsrcName == "out"

        assert len(edges) == 2
        assert edges[0].source == nodes[0].id
        assert edges[0].target == nodes[1].id
        assert edges[1].source == nodes[1].id
        assert edges[1].target == nodes[2].id


def test_car(mock_discovery: Any, mock_auth: Any) -> None:  # type: ignore[explicit-any]
    """Test car builds correct TransformNode and returns output metadata."""
    client, _ = _make_client_with_dataset(mock_discovery, mock_auth)
    mock_output = MagicMock()

    with (
        patch.object(client, "_execute_protocol") as mock_exec,
        patch.object(client, "link_data_file", return_value=mock_output),
    ):
        res = client.car("input_ds", "car_out.xdat")

        assert res == mock_output
        mock_exec.assert_called_once()
        nodes, _edges = mock_exec.call_args[0]

        assert len(nodes) == 3
        assert nodes[1].type == common_pb2.CAR_REF
        assert "CAR car_out.xdat" in mock_exec.call_args[1]["desc"]


def test_car_graph_structure(mock_discovery: Any, mock_auth: Any) -> None:  # type: ignore[explicit-any]
    """Test car produces correct Source -> CAR_REF -> Sink graph."""
    client, _ = _make_client_with_dataset(mock_discovery, mock_auth)

    with (
        patch.object(client, "_execute_protocol") as mock_exec,
        patch.object(client, "link_data_file", return_value=MagicMock()),
    ):
        client.car("input_ds", "car_out.xdat")

        nodes, edges = mock_exec.call_args[0]

        assert nodes[0].type == common_pb2.SOURCE
        assert nodes[0].datasourceSinkParams.id == "input_ds"
        assert nodes[1].type == common_pb2.CAR_REF
        assert nodes[2].type == common_pb2.SINK
        assert nodes[2].datasourceSinkParams.dsrcName == "car_out"

        assert len(edges) == 2
        assert edges[0].source == nodes[0].id
        assert edges[0].target == nodes[1].id
        assert edges[1].source == nodes[1].id
        assert edges[1].target == nodes[2].id


def test_virtual_ref(mock_discovery: Any, mock_auth: Any) -> None:  # type: ignore[explicit-any]
    """Test virtual_ref builds correct TransformNode with refNtvChanIdx."""
    client, _ = _make_client_with_dataset(mock_discovery, mock_auth)
    mock_output = MagicMock()

    with (
        patch.object(client, "_execute_protocol") as mock_exec,
        patch.object(client, "link_data_file", return_value=mock_output),
    ):
        res = client.virtual_ref("input_ds", ref_channel=3, output_path="vref_out.xdat")

        assert res == mock_output
        nodes, _edges = mock_exec.call_args[0]

        assert len(nodes) == 3
        assert nodes[1].type == common_pb2.VIRTUAL_REF
        assert nodes[1].virtualRefParams.refNtvChanIdx == 3


def test_paired_ref(mock_discovery: Any, mock_auth: Any) -> None:  # type: ignore[explicit-any]
    """Test paired_ref builds correct TransformNode with both channel indices."""
    client, _ = _make_client_with_dataset(mock_discovery, mock_auth)
    mock_output = MagicMock()

    with (
        patch.object(client, "_execute_protocol") as mock_exec,
        patch.object(client, "link_data_file", return_value=mock_output),
    ):
        res = client.paired_ref("input_ds", ref_channel=1, target_channel=4, output_path="pref_out.xdat")

        assert res == mock_output
        nodes, _edges = mock_exec.call_args[0]

        assert len(nodes) == 3
        assert nodes[1].type == common_pb2.PAIRED_REF
        assert nodes[1].pairedRefParams.refNtvChanIdx == 1
        assert nodes[1].pairedRefParams.targetNtvChanIdx == 4


def test_paired_ref_channel_args(mock_discovery: Any, mock_auth: Any) -> None:  # type: ignore[explicit-any]
    """Test paired_ref correctly maps ref=2, target=5."""
    client, _ = _make_client_with_dataset(mock_discovery, mock_auth)

    with (
        patch.object(client, "_execute_protocol") as mock_exec,
        patch.object(client, "link_data_file", return_value=MagicMock()),
    ):
        client.paired_ref("input_ds", ref_channel=2, target_channel=5, output_path="pref_out.xdat")

        nodes, _edges = mock_exec.call_args[0]

        assert nodes[1].pairedRefParams.refNtvChanIdx == 2
        assert nodes[1].pairedRefParams.targetNtvChanIdx == 5


# --- Protocol Management Tests ---


def test_apply_protocol(mock_discovery: Any, mock_auth: Any) -> None:  # type: ignore[explicit-any]
    """Test apply_protocol consumes the progress stream."""
    client = CurateClient()

    def progress_stream(*args: Any, **kwargs: Any) -> Generator[common_pb2.ApplyProtocolProgress, None, None]:  # type: ignore[explicit-any]
        yield common_pb2.ApplyProtocolProgress(fracComplete=0.5)
        yield common_pb2.ApplyProtocolProgress(fracComplete=1.0)

    mock_apply = MagicMock(return_value=progress_stream())

    with (
        patch("radiens_core.curate_client.apply_protocol_proto", mock_apply),
        patch("radiens_core.curate_client.tqdm") as mock_tqdm,
    ):
        pbar = mock_tqdm.return_value.__enter__.return_value
        client.apply_protocol("my-protocol-id")

        mock_apply.assert_called_once()
        req = mock_apply.call_args[0][2]
        assert req.protocolID == "my-protocol-id"
        assert pbar.update.call_count >= 1


def test_get_protocol(mock_discovery: Any, mock_auth: Any) -> None:  # type: ignore[explicit-any]
    """Test get_protocol returns a ProtocolSpec."""
    client = CurateClient()

    mock_proto = common_pb2.Protocol(
        id="proto-abc",
        transformNodes=[
            common_pb2.TransformNode(id="n1"),
            common_pb2.TransformNode(id="n2"),
        ],
    )
    mock_get = MagicMock(return_value=mock_proto)

    with patch("radiens_core.curate_client.get_protocol_proto", mock_get):
        result = client.get_protocol("proto-abc")

    assert isinstance(result, ProtocolSpec)
    assert result.id == "proto-abc"
    assert result.node_count == 2
    mock_get.assert_called_once()
    assert mock_get.call_args[0][2] == "proto-abc"


def test_get_all_protocols(mock_discovery: Any, mock_auth: Any) -> None:  # type: ignore[explicit-any]
    """Test get_all_protocols returns a list of ProtocolSpec."""
    client = CurateClient()

    mock_reply = common_pb2.GetAllProtocolsReply(
        protocols=[
            common_pb2.Protocol(id="p1", transformNodes=[common_pb2.TransformNode(id="n1")]),
            common_pb2.Protocol(
                id="p2",
                transformNodes=[common_pb2.TransformNode(id="n2"), common_pb2.TransformNode(id="n3")],
            ),
        ]
    )
    mock_get_all = MagicMock(return_value=mock_reply)

    with patch("radiens_core.curate_client.get_all_protocols_proto", mock_get_all):
        result = client.get_all_protocols()

    assert len(result) == 2
    assert result[0].id == "p1"
    assert result[0].node_count == 1
    assert result[1].id == "p2"
    assert result[1].node_count == 2


# --- Bulk Transform Tests ---


def _make_bulk_sources(
    client: CurateClient,
) -> list[MagicMock]:
    """Register two mock datasets and return their mocks."""
    metas = []
    for i in range(2):
        m = MagicMock()
        m.id = f"ds_{i}"
        m.base_name = f"recording_{i}"
        m.path = f"/data/dir_{i}"
        m.file_type = RadiensFileType.XDAT
        client._datasets[m.id] = m
        metas.append(m)
    return metas


def test_bulk_highpass_node_types(mock_discovery: Any, mock_auth: Any) -> None:  # type: ignore[explicit-any]
    """Test bulk_highpass builds BULK_SOURCE/BULK_SINK nodes with correct params."""
    client = CurateClient()
    _make_bulk_sources(client)

    with (
        patch("radiens_core.curate_client.set_protocol_proto") as mock_set,
        patch("radiens_core.curate_client.apply_protocol_proto") as mock_apply,
        patch("radiens_core.curate_client.tqdm") as mock_tqdm,
    ):
        mock_set.return_value = common_pb2.Protocol(transformNodes=[])
        mock_tqdm.return_value.__enter__.return_value = MagicMock()

        def _empty_stream(*args: Any, **kwargs: Any) -> Generator[common_pb2.ApplyProtocolProgress, None, None]:  # type: ignore[explicit-any]
            yield common_pb2.ApplyProtocolProgress(fracComplete=1.0)

        mock_apply.return_value = _empty_stream()

        client.bulk_highpass(["ds_0", "ds_1"], 300.0, "/out", "_hp")

    # Verify nodes sent to set_protocol_proto
    protocol = mock_set.call_args[0][2]
    nodes = list(protocol.transformNodes)

    assert nodes[0].type == common_pb2.BULK_SOURCE
    src_params = list(nodes[0].bulkDataSourceParams.srcParams)
    assert len(src_params) == 2
    assert src_params[0].id == "ds_0"
    assert src_params[1].id == "ds_1"

    assert nodes[1].type == common_pb2.HIGHPASS_FILTER
    assert nodes[1].highLowPassParams.frequency == 300.0

    assert nodes[2].type == common_pb2.BULK_SINK
    assert nodes[2].bulkDataSinkParams.suffix == "_hp"
    assert nodes[2].bulkDataSinkParams.path == "/out"

    # Verify edges
    edges = list(protocol.transformEdges)
    assert len(edges) == 2
    assert edges[0].source == nodes[0].id
    assert edges[0].target == nodes[1].id
    assert edges[1].source == nodes[1].id
    assert edges[1].target == nodes[2].id


# --- set_protocol Tests ---


def test_set_protocol_returns_spec(mock_discovery: Any, mock_auth: Any) -> None:  # type: ignore[explicit-any]
    """set_protocol converts domain nodes/edges, submits to server, returns ProtocolSpec."""
    client = CurateClient()

    src = TransformNode(
        id="src",
        type=TransformNodeType.SOURCE,
        params=SourceParams(dataset_id="ds1", base_name="rec", path="/data", file_type=RadiensFileType.XDAT),
    )
    hp = TransformNode(
        id="hp",
        type=TransformNodeType.HIGHPASS_FILTER,
        params=HighLowPassParams(frequency=300.0),
    )
    snk = TransformNode(
        id="snk",
        type=TransformNodeType.SINK,
        params=SinkParams(base_name="out", path="/output", file_type=RadiensFileType.XDAT),
    )
    edges = [
        TransformEdge(id="e1", source="src", target="hp"),
        TransformEdge(id="e2", source="hp", target="snk"),
    ]

    mock_validated = common_pb2.Protocol(
        id="proto-abc",
        transformNodes=[
            common_pb2.TransformNode(id="src", invalid=False),
            common_pb2.TransformNode(id="hp", invalid=False),
            common_pb2.TransformNode(id="snk", invalid=False),
        ],
    )
    mock_set = MagicMock(return_value=mock_validated)

    with patch("radiens_core.curate_client.set_protocol_proto", mock_set):
        result = client.set_protocol([src, hp, snk], edges)

    assert isinstance(result, ProtocolSpec)
    assert result.id == "proto-abc"
    assert result.node_count == 3

    # Verify the proto sent to the server has the right structure
    sent_proto = mock_set.call_args[0][2]
    sent_nodes = list(sent_proto.transformNodes)
    assert len(sent_nodes) == 3
    assert sent_nodes[0].id == "src"
    assert sent_nodes[0].type == common_pb2.SOURCE
    assert sent_nodes[1].type == common_pb2.HIGHPASS_FILTER
    assert sent_nodes[1].highLowPassParams.frequency == 300.0
    assert sent_nodes[2].type == common_pb2.SINK

    sent_edges = list(sent_proto.transformEdges)
    assert len(sent_edges) == 2
    assert sent_edges[0].source == "src"
    assert sent_edges[0].target == "hp"


def test_set_protocol_raises_on_invalid_node(mock_discovery: Any, mock_auth: Any) -> None:  # type: ignore[explicit-any]
    """set_protocol raises RadiensError when the server marks a node invalid."""
    from radiens_core.exceptions import RadiensError

    client = CurateClient()

    node = TransformNode.car()
    mock_validated = common_pb2.Protocol(
        id="bad-proto",
        transformNodes=[
            common_pb2.TransformNode(id=node.id, invalid=True, errorMessage="missing source"),
        ],
    )
    mock_set = MagicMock(return_value=mock_validated)

    with (
        patch("radiens_core.curate_client.set_protocol_proto", mock_set),
        pytest.raises(RadiensError, match="missing source"),
    ):
        client.set_protocol([node], [])


def test_set_protocol_custom_id(mock_discovery: Any, mock_auth: Any) -> None:  # type: ignore[explicit-any]
    """set_protocol uses the caller-supplied protocol_id."""
    client = CurateClient()

    node = TransformNode(id="n1", type=TransformNodeType.CAR_REF, params=CARParams())
    mock_validated = common_pb2.Protocol(
        id="my-custom-id",
        transformNodes=[common_pb2.TransformNode(id="n1", invalid=False)],
    )
    mock_set = MagicMock(return_value=mock_validated)

    with patch("radiens_core.curate_client.set_protocol_proto", mock_set):
        result = client.set_protocol([node], [], protocol_id="my-custom-id")

    sent_proto = mock_set.call_args[0][2]
    assert sent_proto.id == "my-custom-id"
    assert result.id == "my-custom-id"


def test_set_protocol_auto_generates_id(mock_discovery: Any, mock_auth: Any) -> None:  # type: ignore[explicit-any]
    """set_protocol generates a UUID when protocol_id is omitted."""
    client = CurateClient()

    node = TransformNode(id="n1", type=TransformNodeType.CAR_REF, params=CARParams())
    mock_set = MagicMock(
        return_value=common_pb2.Protocol(
            id="server-id",
            transformNodes=[common_pb2.TransformNode(id="n1", invalid=False)],
        )
    )

    with patch("radiens_core.curate_client.set_protocol_proto", mock_set):
        client.set_protocol([node], [])

    sent_proto = mock_set.call_args[0][2]
    # A UUID was generated — must be non-empty and not a fixed string
    assert len(sent_proto.id) > 0



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

from . import biointerface_pb2
import builtins
import collections.abc
from . import common_pb2
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import typing

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class RadiensHealthcheckRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ISINCLUDEWAREHOUSE_FIELD_NUMBER: builtins.int
    ISDETAILED_FIELD_NUMBER: builtins.int
    isIncludeWarehouse: builtins.bool
    isDetailed: builtins.bool
    def __init__(
        self,
        *,
        isIncludeWarehouse: builtins.bool | None = ...,
        isDetailed: builtins.bool | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["isDetailed", b"isDetailed", "isIncludeWarehouse", b"isIncludeWarehouse"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["isDetailed", b"isDetailed", "isIncludeWarehouse", b"isIncludeWarehouse"]) -> None: ...

global___RadiensHealthcheckRequest = RadiensHealthcheckRequest

@typing.final
class RadiensHealthcheckSpec(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ERRMSG_FIELD_NUMBER: builtins.int
    OUTFITTER_FIELD_NUMBER: builtins.int
    errMsg: builtins.str
    @property
    def outfitter(self) -> common_pb2.WarehouseHealthcheckSpec: ...
    def __init__(
        self,
        *,
        errMsg: builtins.str | None = ...,
        outfitter: common_pb2.WarehouseHealthcheckSpec | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["errMsg", b"errMsg", "outfitter", b"outfitter"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["errMsg", b"errMsg", "outfitter", b"outfitter"]) -> None: ...

global___RadiensHealthcheckSpec = RadiensHealthcheckSpec

@typing.final
class RadiensClientId(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PID_FIELD_NUMBER: builtins.int
    pid: builtins.int
    """the pid of the process that's establishing a long-running session with radiensserver"""
    def __init__(
        self,
        *,
        pid: builtins.int | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["pid", b"pid"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["pid", b"pid"]) -> None: ...

global___RadiensClientId = RadiensClientId

@typing.final
class SummaSessionsReply(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class SummaCurrentSession(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        SUMMAID_FIELD_NUMBER: builtins.int
        DSOURCEIDS_FIELD_NUMBER: builtins.int
        summaID: builtins.str
        @property
        def dsourceIDs(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
        def __init__(
            self,
            *,
            summaID: builtins.str | None = ...,
            dsourceIDs: collections.abc.Iterable[builtins.str] | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["summaID", b"summaID"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["dsourceIDs", b"dsourceIDs", "summaID", b"summaID"]) -> None: ...

    SESSION_FIELD_NUMBER: builtins.int
    @property
    def session(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___SummaSessionsReply.SummaCurrentSession]: ...
    def __init__(
        self,
        *,
        session: collections.abc.Iterable[global___SummaSessionsReply.SummaCurrentSession] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["session", b"session"]) -> None: ...

global___SummaSessionsReply = SummaSessionsReply

@typing.final
class SummaAnalysisRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SUMMAID_FIELD_NUMBER: builtins.int
    DSOURCEIDS_FIELD_NUMBER: builtins.int
    PKTDURSEC_FIELD_NUMBER: builtins.int
    REQPART_FIELD_NUMBER: builtins.int
    summaID: builtins.str
    pktDurSec: builtins.float
    reqPart: builtins.str
    @property
    def dsourceIDs(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    def __init__(
        self,
        *,
        summaID: builtins.str | None = ...,
        dsourceIDs: collections.abc.Iterable[builtins.str] | None = ...,
        pktDurSec: builtins.float | None = ...,
        reqPart: builtins.str | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["pktDurSec", b"pktDurSec", "reqPart", b"reqPart", "summaID", b"summaID"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["dsourceIDs", b"dsourceIDs", "pktDurSec", b"pktDurSec", "reqPart", b"reqPart", "summaID", b"summaID"]) -> None: ...

global___SummaAnalysisRequest = SummaAnalysisRequest

@typing.final
class SummaAnalysisReply(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SUMMAID_FIELD_NUMBER: builtins.int
    ALLDSOURCEIDS_FIELD_NUMBER: builtins.int
    STATE_FIELD_NUMBER: builtins.int
    ALLDSOURCESPEC_FIELD_NUMBER: builtins.int
    AGGSTATS_FIELD_NUMBER: builtins.int
    REQAVAILDSOURCESTATS_FIELD_NUMBER: builtins.int
    REQAVAILDSOURCEIDS_FIELD_NUMBER: builtins.int
    summaID: builtins.str
    """returns nil if no summa sessions, otherwise always returns"""
    @property
    def allDsourceIDs(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """always returns all dsourceIDs"""

    @property
    def state(self) -> global___SummaState:
        """always returns"""

    @property
    def allDsourceSpec(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___SummaDatasourceSpec]:
        """always returns all dsourceSpec"""

    @property
    def aggStats(self) -> global___SummaAggregateStats: ...
    @property
    def reqAvailDsourceStats(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___SummaDatasourceAnalysisPkg]: ...
    @property
    def reqAvailDsourceIDs(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """returns available req DsourceIDs"""

    def __init__(
        self,
        *,
        summaID: builtins.str | None = ...,
        allDsourceIDs: collections.abc.Iterable[builtins.str] | None = ...,
        state: global___SummaState | None = ...,
        allDsourceSpec: collections.abc.Iterable[global___SummaDatasourceSpec] | None = ...,
        aggStats: global___SummaAggregateStats | None = ...,
        reqAvailDsourceStats: collections.abc.Iterable[global___SummaDatasourceAnalysisPkg] | None = ...,
        reqAvailDsourceIDs: collections.abc.Iterable[builtins.str] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["aggStats", b"aggStats", "state", b"state", "summaID", b"summaID"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["aggStats", b"aggStats", "allDsourceIDs", b"allDsourceIDs", "allDsourceSpec", b"allDsourceSpec", "reqAvailDsourceIDs", b"reqAvailDsourceIDs", "reqAvailDsourceStats", b"reqAvailDsourceStats", "state", b"state", "summaID", b"summaID"]) -> None: ...

global___SummaAnalysisReply = SummaAnalysisReply

@typing.final
class SummaState(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SUMMAID_FIELD_NUMBER: builtins.int
    ISCOMPLETE_FIELD_NUMBER: builtins.int
    FRACCOMPLETE_FIELD_NUMBER: builtins.int
    MSG_FIELD_NUMBER: builtins.int
    DSRCWALLTIMESTART_FIELD_NUMBER: builtins.int
    DSRCWALLTIMEEND_FIELD_NUMBER: builtins.int
    ANALYSISWALLTIMESTART_FIELD_NUMBER: builtins.int
    ANALYSISELAPSEDSEC_FIELD_NUMBER: builtins.int
    PACKETDURSEC_FIELD_NUMBER: builtins.int
    summaID: builtins.str
    isComplete: builtins.bool
    fracComplete: builtins.float
    msg: builtins.str
    dsrcWalltimeStart: builtins.str
    dsrcWalltimeEnd: builtins.str
    analysisWallTimeStart: builtins.str
    analysisElapsedSec: builtins.float
    packetDurSec: builtins.float
    def __init__(
        self,
        *,
        summaID: builtins.str | None = ...,
        isComplete: builtins.bool | None = ...,
        fracComplete: builtins.float | None = ...,
        msg: builtins.str | None = ...,
        dsrcWalltimeStart: builtins.str | None = ...,
        dsrcWalltimeEnd: builtins.str | None = ...,
        analysisWallTimeStart: builtins.str | None = ...,
        analysisElapsedSec: builtins.float | None = ...,
        packetDurSec: builtins.float | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["analysisElapsedSec", b"analysisElapsedSec", "analysisWallTimeStart", b"analysisWallTimeStart", "dsrcWalltimeEnd", b"dsrcWalltimeEnd", "dsrcWalltimeStart", b"dsrcWalltimeStart", "fracComplete", b"fracComplete", "isComplete", b"isComplete", "msg", b"msg", "packetDurSec", b"packetDurSec", "summaID", b"summaID"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["analysisElapsedSec", b"analysisElapsedSec", "analysisWallTimeStart", b"analysisWallTimeStart", "dsrcWalltimeEnd", b"dsrcWalltimeEnd", "dsrcWalltimeStart", b"dsrcWalltimeStart", "fracComplete", b"fracComplete", "isComplete", b"isComplete", "msg", b"msg", "packetDurSec", b"packetDurSec", "summaID", b"summaID"]) -> None: ...

global___SummaState = SummaState

@typing.final
class SummaAggregateStats(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    METRICID_FIELD_NUMBER: builtins.int
    HIST_FIELD_NUMBER: builtins.int
    @property
    def metricID(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[common_pb2.KpiMetricID]: ...
    @property
    def hist(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[common_pb2.Histogram]: ...
    def __init__(
        self,
        *,
        metricID: collections.abc.Iterable[common_pb2.KpiMetricID] | None = ...,
        hist: collections.abc.Iterable[common_pb2.Histogram] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["hist", b"hist", "metricID", b"metricID"]) -> None: ...

global___SummaAggregateStats = SummaAggregateStats

@typing.final
class SummaDatasourceAnalysisPkg(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    METRICID_FIELD_NUMBER: builtins.int
    HIST_FIELD_NUMBER: builtins.int
    RECS_FIELD_NUMBER: builtins.int
    HISTRECS_FIELD_NUMBER: builtins.int
    NTVIDXS_FIELD_NUMBER: builtins.int
    @property
    def metricID(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[common_pb2.KpiMetricID]: ...
    @property
    def hist(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[common_pb2.Histogram]: ...
    @property
    def recs(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[common_pb2.DenseMatrix]: ...
    @property
    def histRecs(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[common_pb2.Histogram]: ...
    @property
    def ntvIdxs(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    def __init__(
        self,
        *,
        metricID: collections.abc.Iterable[common_pb2.KpiMetricID] | None = ...,
        hist: collections.abc.Iterable[common_pb2.Histogram] | None = ...,
        recs: collections.abc.Iterable[common_pb2.DenseMatrix] | None = ...,
        histRecs: collections.abc.Iterable[common_pb2.Histogram] | None = ...,
        ntvIdxs: collections.abc.Iterable[builtins.int] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["hist", b"hist", "histRecs", b"histRecs", "metricID", b"metricID", "ntvIdxs", b"ntvIdxs", "recs", b"recs"]) -> None: ...

global___SummaDatasourceAnalysisPkg = SummaDatasourceAnalysisPkg

@typing.final
class SummaDatasourceSpec(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SG_FIELD_NUMBER: builtins.int
    TR_FIELD_NUMBER: builtins.int
    NISPEC_FIELD_NUMBER: builtins.int
    @property
    def sG(self) -> common_pb2.SignalGroup: ...
    @property
    def tR(self) -> common_pb2.TimeRange: ...
    @property
    def niSpec(self) -> biointerface_pb2.SpikesSpecReply: ...
    def __init__(
        self,
        *,
        sG: common_pb2.SignalGroup | None = ...,
        tR: common_pb2.TimeRange | None = ...,
        niSpec: biointerface_pb2.SpikesSpecReply | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["niSpec", b"niSpec", "sG", b"sG", "tR", b"tR"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["niSpec", b"niSpec", "sG", b"sG", "tR", b"tR"]) -> None: ...

global___SummaDatasourceSpec = SummaDatasourceSpec



"""Tests for restart converters."""

from radiens_core._converters._restart import restart_request_to_proto
from radiens_core.models.status import BackboneMode, RestartRequest


def test_restart_request_to_proto() -> None:
    """Test conversion of RestartRequest to proto."""
    # Test different backbone modes
    request_smartbox = RestartRequest(mode=BackboneMode.SMARTBOX_PRO)
    proto_smartbox = restart_request_to_proto(request_smartbox)
    assert proto_smartbox.mode == 0  # SMARTBOX_PRO proto value

    # Test simulation mode
    request_sim = RestartRequest(mode=BackboneMode.SMARTBOX_SIM_GEN_SINE)
    proto_sim = restart_request_to_proto(request_sim)
    assert proto_sim.mode == 1  # SMARTBOX_SIM_GEN_SINE proto value

    # Test XDAQ mode
    request_xdaq = RestartRequest(mode=BackboneMode.XDAQ_ONE_REC)
    proto_xdaq = restart_request_to_proto(request_xdaq)
    assert proto_xdaq.mode == 15  # XDAQ_ONE_REC proto value
